#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018, 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""MITIFF writer objects for creating MITIFF files from `Dataset` objects."""

import logging

import dask
import numpy as np

from satpy.dataset import DataID, DataQuery
from satpy.writers import ImageWriter, get_enhanced_image

IMAGEDESCRIPTION = 270

LOG = logging.getLogger(__name__)

KELVIN_TO_CELSIUS = -273.15


def _adjust_kwargs(dataset, kwargs):
    if 'platform_name' not in kwargs:
        kwargs['platform_name'] = dataset.attrs['platform_name']
    if 'name' not in kwargs:
        kwargs['name'] = dataset.attrs['name']
    if 'start_time' not in kwargs:
        kwargs['start_time'] = dataset.attrs['start_time']
    if 'sensor' not in kwargs:
        kwargs['sensor'] = dataset.attrs['sensor']
    # Sensor attrs could be set. MITIFFs needing to handle sensor can only have one sensor
    # Assume the first value of set as the sensor.
    if isinstance(kwargs['sensor'], set):
        LOG.warning('Sensor is set, will use the first value: %s', kwargs['sensor'])
        kwargs['sensor'] = (list(kwargs['sensor']))[0]


class MITIFFWriter(ImageWriter):
    """Writer to produce MITIFF image files."""

    def __init__(self, name=None, tags=None, **kwargs):
        """Initialize reader with tag and other configuration information."""
        ImageWriter.__init__(self, name=name, default_config_filename="writers/mitiff.yaml", **kwargs)
        self.tags = self.info.get("tags", None) if tags is None else tags
        if self.tags is None:
            self.tags = {}
        elif not isinstance(self.tags, dict):
            # if it's coming from a config file
            self.tags = dict(tuple(x.split("=")) for x in self.tags.split(","))

        self.mitiff_config = {}
        self.translate_channel_name = {}
        self.channel_order = {}
        self.palette = False
        self.sensor = None

    def save_image(self):
        """Save dataset as an image array."""
        raise NotImplementedError("save_image mitiff is not implemented.")

    def save_dataset(self, dataset, filename=None, fill_value=None,
                     compute=True, **kwargs):
        """Save single dataset as mitiff file."""
        LOG.debug("Starting in mitiff save_dataset ... ")

        def _delayed_create(dataset):
            try:
                if 'palette' in kwargs:
                    self.palette = kwargs['palette']
                _adjust_kwargs(dataset, kwargs)

                try:
                    self.mitiff_config[kwargs['sensor']] = dataset.attrs['metadata_requirements']['config']
                    self.channel_order[kwargs['sensor']] = dataset.attrs['metadata_requirements']['order']
                    self.file_pattern = dataset.attrs['metadata_requirements']['file_pattern']
                except KeyError:
                    # For some mitiff products this info is needed, for others not.
                    # If needed you should know how to fix this
                    pass

                try:
                    self.translate_channel_name[kwargs['sensor']] = \
                        dataset.attrs['metadata_requirements']['translate']
                except KeyError:
                    # For some mitiff products this info is needed, for others not.
                    # If needed you should know how to fix this
                    pass

                image_description = self._make_image_description(dataset, **kwargs)
                gen_filename = filename or self.get_filename(**dataset.attrs)
                LOG.info("Saving mitiff to: %s ...", gen_filename)
                self._save_datasets_as_mitiff(dataset, image_description,
                                              gen_filename, **kwargs)
            except (KeyError, ValueError, RuntimeError):
                raise

        delayed = dask.delayed(_delayed_create)(dataset)

        if compute:
            return delayed.compute()
        return delayed

    def save_datasets(self, datasets, filename=None, fill_value=None,
                      compute=True, **kwargs):
        """Save all datasets to one or more files."""
        LOG.debug("Starting in mitiff save_datasets ... ")

        def _delayed_create(datasets):
            dataset = datasets[0]

            try:
                _adjust_kwargs(dataset, kwargs)

                try:
                    self.mitiff_config[kwargs['sensor']] = dataset.attrs['metadata_requirements']['config']
                    translate = dataset.attrs['metadata_requirements']['translate']
                    self.translate_channel_name[kwargs['sensor']] = translate
                    self.channel_order[kwargs['sensor']] = dataset.attrs['metadata_requirements']['order']
                    self.file_pattern = dataset.attrs['metadata_requirements']['file_pattern']
                except KeyError:
                    # For some mitiff products this info is needed, for others not.
                    # If needed you should know how to fix this
                    pass

                image_description = self._make_image_description(datasets, **kwargs)
                LOG.debug("File pattern %s", self.file_pattern)
                if isinstance(datasets, list):
                    kwargs['start_time'] = dataset.attrs['start_time']
                else:
                    kwargs['start_time'] = datasets.attrs['start_time']
                gen_filename = filename or self.get_filename(**kwargs)
                LOG.info("Saving mitiff to: %s ...", gen_filename)
                self._save_datasets_as_mitiff(datasets, image_description, gen_filename, **kwargs)
            except (KeyError, ValueError, RuntimeError):
                raise

        delayed = dask.delayed(_delayed_create)(datasets)
        LOG.debug("About to call delayed compute ...")
        if compute:
            return delayed.compute()
        return delayed

    def _make_channel_list(self, datasets, **kwargs):
        channels = []
        try:
            if self.channel_order:
                channels = self._reorder_channels(datasets, **kwargs)
            elif self.palette:
                if 'palette_channel_name' in kwargs:
                    channels.append(kwargs['palette_channel_name'].upper())
                else:
                    LOG.error("Is palette but can not find palette_channel_name to name the dataset")
            else:
                for ch in range(len(datasets)):
                    channels.append(ch + 1)
        except KeyError:
            for ch in range(len(datasets)):
                channels.append(ch + 1)
        return channels

    def _reorder_channels(self, datasets, **kwargs):
        channels = []
        for cn in self.channel_order[kwargs['sensor']]:
            for ch, ds in enumerate(datasets):
                if isinstance(ds.attrs['prerequisites'][ch], (DataQuery, DataID)):
                    if ds.attrs['prerequisites'][ch]['name'] == cn:
                        channels.append(
                            ds.attrs['prerequisites'][ch]['name'])
                        break
                else:
                    if ds.attrs['prerequisites'][ch] == cn:
                        channels.append(
                            ds.attrs['prerequisites'][ch])
                        break
        return channels

    def _channel_names(self, channels, cns, **kwargs):
        _image_description = ""
        for ch in channels:
            try:
                _image_description += str(
                    self.mitiff_config[kwargs['sensor']][cns.get(ch, ch)]['alias'])
            except KeyError:
                _image_description += str(ch)
            _image_description += ' '

        # Replace last char(space) with \n
        _image_description = _image_description[:-1]
        _image_description += '\n'

        return _image_description

    def _add_sizes(self, datasets, first_dataset):
        _image_description = ' Xsize: '
        if isinstance(datasets, list):
            _image_description += str(first_dataset.sizes['x']) + '\n'
        else:
            _image_description += str(datasets.sizes['x']) + '\n'

        _image_description += ' Ysize: '
        if isinstance(datasets, list):
            _image_description += str(first_dataset.sizes['y']) + '\n'
        else:
            _image_description += str(datasets.sizes['y']) + '\n'

        return _image_description

    def _add_proj4_string(self, datasets, first_dataset):
        proj4_string = " Proj string: "

        if isinstance(datasets, list):
            area = first_dataset.attrs['area']
        else:
            area = datasets.attrs['area']
        # Use pyproj's CRS object to get a valid EPSG code if possible
        # only in newer pyresample versions with pyproj 2.0+ installed
        if hasattr(area, 'crs') and area.crs.to_epsg() is not None:
            proj4_string += "+init=EPSG:{}".format(area.crs.to_epsg())
        else:
            proj4_string += area.proj_str

        x_0 = 0
        y_0 = 0
        # FUTURE: Use pyproj 2.0+ to convert EPSG to PROJ4 if possible
        proj4_string, x_0 = self._convert_epsg_to_proj(proj4_string, x_0)

        if 'geos' in proj4_string:
            proj4_string = proj4_string.replace("+sweep=x ", "")
            if '+a=6378137.0 +b=6356752.31414' in proj4_string:
                proj4_string = proj4_string.replace("+a=6378137.0 +b=6356752.31414",
                                                    "+ellps=WGS84")
        if '+units=m' in proj4_string:
            proj4_string = proj4_string.replace("+units=m", "+units=km")

        if not any(datum in proj4_string for datum in ['datum', 'towgs84']):
            proj4_string += ' +towgs84=0,0,0'

        if 'units' not in proj4_string:
            proj4_string += ' +units=km'

        proj4_string = self._append_projection_center(proj4_string, datasets, first_dataset, x_0, y_0)
        LOG.debug("proj4_string: %s", proj4_string)
        proj4_string += '\n'

        return proj4_string

    def _append_projection_center(self, proj4_string, datasets, first_dataset, x_0, y_0):
        if isinstance(datasets, list):
            dataset = first_dataset
        else:
            dataset = datasets
        if 'x_0' not in proj4_string:
            proj4_string += ' +x_0=%.6f' % (
                    (-dataset.attrs['area'].area_extent[0] +
                     dataset.attrs['area'].pixel_size_x) + x_0)
            proj4_string += ' +y_0=%.6f' % (
                    (-dataset.attrs['area'].area_extent[1] +
                     dataset.attrs['area'].pixel_size_y) + y_0)
        elif '+x_0=0' in proj4_string and '+y_0=0' in proj4_string:
            proj4_string = proj4_string.replace("+x_0=0", '+x_0=%.6f' % (
                    (-dataset.attrs['area'].area_extent[0] +
                     dataset.attrs['area'].pixel_size_x) + x_0))
            proj4_string = proj4_string.replace("+y_0=0", '+y_0=%.6f' % (
                    (-dataset.attrs['area'].area_extent[1] +
                     dataset.attrs['area'].pixel_size_y) + y_0))
        return proj4_string

    def _convert_epsg_to_proj(self, proj4_string, x_0):
        if 'EPSG:32631' in proj4_string:
            proj4_string = proj4_string.replace("+init=EPSG:32631",
                                                "+proj=etmerc +lat_0=0 +lon_0=3 +k=0.9996 +ellps=WGS84 +datum=WGS84")
            x_0 = 500000
        elif 'EPSG:32632' in proj4_string:
            proj4_string = proj4_string.replace("+init=EPSG:32632",
                                                "+proj=etmerc +lat_0=0 +lon_0=9 +k=0.9996 +ellps=WGS84 +datum=WGS84")
            x_0 = 500000
        elif 'EPSG:32633' in proj4_string:
            proj4_string = proj4_string.replace("+init=EPSG:32633",
                                                "+proj=etmerc +lat_0=0 +lon_0=15 +k=0.9996 +ellps=WGS84 +datum=WGS84")
            x_0 = 500000
        elif 'EPSG:32634' in proj4_string:
            proj4_string = proj4_string.replace("+init=EPSG:32634",
                                                "+proj=etmerc +lat_0=0 +lon_0=21 +k=0.9996 +ellps=WGS84 +datum=WGS84")
            x_0 = 500000
        elif 'EPSG:32635' in proj4_string:
            proj4_string = proj4_string.replace("+init=EPSG:32635",
                                                "+proj=etmerc +lat_0=0 +lon_0=27 +k=0.9996 +ellps=WGS84 +datum=WGS84")
            x_0 = 500000
        elif 'EPSG' in proj4_string:
            LOG.warning("EPSG used in proj string but not converted. Please add this in code")
        return proj4_string, x_0

    def _add_pixel_sizes(self, datasets, first_dataset):
        _image_description = ""
        if isinstance(datasets, list):
            _image_description += ' Ax: %.6f' % (
                first_dataset.attrs['area'].pixel_size_x / 1000.)
            _image_description += ' Ay: %.6f' % (
                first_dataset.attrs['area'].pixel_size_y / 1000.)
        else:
            _image_description += ' Ax: %.6f' % (
                datasets.attrs['area'].pixel_size_x / 1000.)
            _image_description += ' Ay: %.6f' % (
                datasets.attrs['area'].pixel_size_y / 1000.)

        return _image_description

    def _add_corners(self, datasets, first_dataset):
        # But this ads up to upper left corner of upper left pixel.
        # But need to use the center of the pixel.
        # Therefor use the center of the upper left pixel.
        _image_description = ""
        if isinstance(datasets, list):
            _image_description += ' Bx: %.6f' % (
                first_dataset.attrs['area'].area_extent[0] / 1000. +
                first_dataset.attrs['area'].pixel_size_x / 1000. / 2.)  # LL_x
            _image_description += ' By: %.6f' % (
                first_dataset.attrs['area'].area_extent[3] / 1000. -
                first_dataset.attrs['area'].pixel_size_y / 1000. / 2.)  # UR_y
        else:
            _image_description += ' Bx: %.6f' % (
                datasets.attrs['area'].area_extent[0] / 1000. +
                datasets.attrs['area'].pixel_size_x / 1000. / 2.)  # LL_x
            _image_description += ' By: %.6f' % (
                datasets.attrs['area'].area_extent[3] / 1000. -
                datasets.attrs['area'].pixel_size_y / 1000. / 2.)  # UR_y

        _image_description += '\n'
        return _image_description

    def _add_calibration_datasets(self, ch, datasets, reverse_offset, reverse_scale, decimals):
        _reverse_offset = reverse_offset
        _reverse_scale = reverse_scale
        _decimals = decimals
        _table_calibration = ""
        found_calibration = False
        skip_calibration = False
        ds_list = datasets
        if not isinstance(datasets, list) and 'bands' not in datasets.sizes:
            ds_list = [datasets]

        for i, ds in enumerate(ds_list):
            if ('prerequisites' in ds.attrs and
                isinstance(ds.attrs['prerequisites'], list) and
                len(ds.attrs['prerequisites']) >= i + 1 and
                    isinstance(ds.attrs['prerequisites'][i], (DataQuery, DataID))):
                if ds.attrs['prerequisites'][i].get('name') == str(ch):
                    if ds.attrs['prerequisites'][i].get('calibration') == 'RADIANCE':
                        raise NotImplementedError(
                            "Mitiff radiance calibration not implemented.")
                    # _table_calibration += ', Radiance, '
                    # _table_calibration += '[W/m²/µm/sr]'
                    # _decimals = 8
                    elif ds.attrs['prerequisites'][i].get('calibration') == 'brightness_temperature':
                        found_calibration = True
                        _table_calibration += ', BT, '
                        _table_calibration += u'\u00B0'  # '\u2103'
                        _table_calibration += u'[C]'

                        _reverse_offset = 255.
                        _reverse_scale = -1.
                        _decimals = 2
                    elif ds.attrs['prerequisites'][i].get('calibration') == 'reflectance':
                        found_calibration = True
                        _table_calibration += ', Reflectance(Albedo), '
                        _table_calibration += '[%]'
                        _decimals = 2
                    else:
                        LOG.warning("Unknown calib type. Must be Radiance, Reflectance or BT.")

                        break
                else:
                    continue
            else:
                _table_calibration = ""
                skip_calibration = True
                break
        if not found_calibration:
            _table_calibration = ""
            skip_calibration = True
            # How to format string by passing the format
            # http://stackoverflow.com/questions/1598579/rounding-decimals-with-new-python-format-function
        return skip_calibration, _table_calibration, _reverse_offset, _reverse_scale, _decimals

    def _add_palette_info(self, datasets, palette_unit, palette_description, **kwargs):
        # mitiff key word for palette interpretion
        _palette = '\n COLOR INFO:\n'
        # mitiff info for the unit of the interpretion
        _palette += ' {}\n'.format(palette_unit)
        # The length of the palette description as needed by mitiff in DIANA
        _palette += ' {}\n'.format(len(palette_description))
        for desc in palette_description:
            _palette += ' {}\n'.format(desc)
        return _palette

    def _add_calibration(self, channels, cns, datasets, **kwargs):

        _table_calibration = ""
        skip_calibration = False
        for ch in channels:

            palette = False
            # Make calibration.
            if palette:
                raise NotImplementedError("Mitiff palette saving is not implemented.")
            else:
                _table_calibration += 'Table_calibration: '
                try:
                    _table_calibration += str(
                        self.mitiff_config[kwargs['sensor']][cns.get(ch, ch)]['alias'])
                except KeyError:
                    _table_calibration += str(ch)

                _reverse_offset = 0.
                _reverse_scale = 1.
                _decimals = 2

                skip_calibration, __table_calibration, _reverse_offset, _reverse_scale, _decimals = \
                    self._add_calibration_datasets(ch, datasets, _reverse_offset, _reverse_scale, _decimals)
                _table_calibration += __table_calibration

                if not skip_calibration:
                    _table_calibration += ', 8, [ '
                    for val in range(0, 256):
                        # Comma separated list of values
                        _table_calibration += '{0:.{1}f} '.format((float(self.mitiff_config[
                            kwargs['sensor']][cns.get(ch, ch)]['min-val']) +
                            ((_reverse_offset + _reverse_scale * val) *
                             (float(self.mitiff_config[kwargs['sensor']][cns.get(ch, ch)]['max-val']) -
                             float(self.mitiff_config[kwargs['sensor']][cns.get(ch, ch)]['min-val']))) / 255.),
                            _decimals)
                        # _table_calibration += '0.00000000 '

                    _table_calibration += ']\n\n'
                else:
                    _table_calibration = ""

        return _table_calibration

    def _make_image_description(self, datasets, **kwargs):
        r"""Generate image description for mitiff.

        Satellite: NOAA 18
        Date and Time: 06:58 31/05-2016
        SatDir: 0
        Channels:   6 In this file: 1-VIS0.63 2-VIS0.86 3(3B)-IR3.7
        4-IR10.8 5-IR11.5 6(3A)-VIS1.6
        Xsize:  4720
        Ysize:  5544
        Map projection: Stereographic
        Proj string: +proj=stere +lon_0=0 +lat_0=90 +lat_ts=60
        +ellps=WGS84 +towgs84=0,0,0 +units=km
        +x_0=2526000.000000 +y_0=5806000.000000
        TrueLat: 60 N
        GridRot: 0
        Xunit:1000 m Yunit: 1000 m
        NPX: 0.000000 NPY: 0.000000
        Ax: 1.000000 Ay: 1.000000 Bx: -2526.000000 By: -262.000000

        Satellite: <satellite name>
        Date and Time: <HH:MM dd/mm-yyyy>
        SatDir: 0
        Channels:   <number of chanels> In this file: <channels names in order>
        Xsize:  <number of pixels x>
        Ysize:  <number of pixels y>
        Map projection: Stereographic
        Proj string: <proj4 string with +x_0 and +y_0 which is the positive
        distance from proj origo
        to the lower left corner of the image data>
        TrueLat: 60 N
        GridRot: 0
        Xunit:1000 m Yunit: 1000 m
        NPX: 0.000000 NPY: 0.000000
        Ax: <pixels size x in km> Ay: <pixel size y in km> Bx: <left corner of
        upper right pixel in km>
        By: <upper corner of upper right pixel in km>


        if palette image write special palette
        if normal channel write table calibration:
        Table_calibration: <channel name>, <calibration type>, [<unit>],
        <no of bits of data>,
        [<calibration values space separated>]\n\n

        """
        translate_platform_name = {'metop01': 'Metop-B',
                                   'metop02': 'Metop-A',
                                   'metop03': 'Metop-C',
                                   'noaa15': 'NOAA-15',
                                   'noaa16': 'NOAA-16',
                                   'noaa17': 'NOAA-17',
                                   'noaa18': 'NOAA-18',
                                   'noaa19': 'NOAA-19'}

        first_dataset = datasets
        if isinstance(datasets, list):
            LOG.debug("Datasets is a list of dataset")
            first_dataset = datasets[0]

        _platform_name = self._get_platform_name(first_dataset, translate_platform_name, kwargs)

        _image_description = ''
        _image_description.encode('utf-8')

        _image_description += ' Satellite: '
        if _platform_name is not None:
            _image_description += _platform_name

        _image_description += '\n'

        _image_description += ' Date and Time: '
        # Select earliest start_time
        first = True
        earliest = 0
        for dataset in datasets:
            if first:
                earliest = dataset.attrs['start_time']
            else:
                if dataset.attrs['start_time'] < earliest:
                    earliest = dataset.attrs['start_time']
            first = False
        LOG.debug("earliest start_time: %s", earliest)
        _image_description += earliest.strftime("%H:%M %d/%m-%Y\n")

        _image_description += ' SatDir: 0\n'

        _image_description += ' Channels: '

        _image_description += self._get_dataset_len(datasets)
        _image_description += ' In this file: '

        channels = self._make_channel_list(datasets, **kwargs)

        try:
            cns = self.translate_channel_name.get(kwargs['sensor'], {})
        except KeyError:
            pass

        _image_description += self._channel_names(channels, cns, **kwargs)

        _image_description += self._add_sizes(datasets, first_dataset)

        _image_description += ' Map projection: Stereographic\n'

        _image_description += self._add_proj4_string(datasets, first_dataset)

        _image_description += ' TrueLat: 60N\n'
        _image_description += ' GridRot: 0\n'

        _image_description += ' Xunit:1000 m Yunit: 1000 m\n'

        _image_description += ' NPX: %.6f' % (0)
        _image_description += ' NPY: %.6f' % (0) + '\n'

        _image_description += self._add_pixel_sizes(datasets, first_dataset)
        _image_description += self._add_corners(datasets, first_dataset)

        if isinstance(datasets, list):
            LOG.debug("Area extent: %s", first_dataset.attrs['area'].area_extent)
        else:
            LOG.debug("Area extent: %s", datasets.attrs['area'].area_extent)

        if self.palette:
            LOG.debug("Doing palette image")
            _image_description += self._add_palette_info(datasets, **kwargs)
        else:
            _image_description += self._add_calibration(channels, cns, datasets, **kwargs)

        return _image_description

    def _get_dataset_len(self, datasets):
        if isinstance(datasets, list):
            LOG.debug("len datasets: %s", len(datasets))
            dataset_len = str(len(datasets))
        elif 'bands' in datasets.sizes:
            LOG.debug("len datasets: %s", datasets.sizes['bands'])
            dataset_len = str(datasets.sizes['bands'])
        elif len(datasets.sizes) == 2:
            LOG.debug("len datasets: 1")
            dataset_len = '1'
        else:
            dataset_len = ""
        return dataset_len

    def _get_platform_name(self, first_dataset, translate_platform_name, kwargs):
        if 'platform_name' in first_dataset.attrs:
            _platform_name = translate_platform_name.get(
                first_dataset.attrs['platform_name'],
                first_dataset.attrs['platform_name'])
        elif 'platform_name' in kwargs:
            _platform_name = translate_platform_name.get(
                kwargs['platform_name'], kwargs['platform_name'])
        else:
            _platform_name = None
        return _platform_name

    def _calibrate_data(self, dataset, calibration, min_val, max_val):
        reverse_offset = 0.
        reverse_scale = 1.
        if calibration == 'brightness_temperature':
            # If data is brightness temperature, the data must be inverted.
            reverse_offset = 255.
            reverse_scale = -1.
            dataset.data += KELVIN_TO_CELSIUS

        # Need to possible translate channels names from satpy to mitiff
        _data = reverse_offset + reverse_scale * ((dataset.data - float(min_val)) /
                                                  (float(max_val) - float(min_val))) * 255.
        return _data.clip(0, 255)

    def _save_as_palette(self, tif, datasets, **kwargs):
        # MITIFF palette has only one data channel
        if len(datasets.dims) == 2:
            LOG.debug("Palette ok with only 2 dimensions. ie only x and y")
            # 3 = Palette color. In this model, a color is described with a single component.
            # The value of the component is used as an index into the red, green and blue curves
            # in the ColorMap field to retrieve an RGB triplet that defines the color. When
            # PhotometricInterpretation=3 is used, ColorMap must be present and SamplesPerPixel must be 1.
            tif.SetField('PHOTOMETRIC', 3)

            # As write_image can not save tiff image as palette, this has to be done basicly
            # ie. all needed tags needs to be set.
            tif.SetField('IMAGEWIDTH', datasets.sizes['x'])
            tif.SetField('IMAGELENGTH', datasets.sizes['y'])
            tif.SetField('BITSPERSAMPLE', 8)
            tif.SetField('COMPRESSION', tif.get_tag_define('deflate'))

            if 'palette_color_map' in kwargs:
                tif.SetField('COLORMAP', kwargs['palette_color_map'])
            else:
                LOG.ERROR("In a mitiff palette image a color map must be provided: palette_color_map is missing.")

            data_type = np.uint8
            # Looks like we need to pass the data to writeencodedstrip as ctypes
            cont_data = np.ascontiguousarray(datasets.data, data_type)
            tif.WriteEncodedStrip(0, cont_data.ctypes.data,
                                  datasets.sizes['x'] * datasets.sizes['y'])
            tif.WriteDirectory()

    def _save_as_enhanced(self, tif, datasets, **kwargs):
        """Save datasets as an enhanced RGB image."""
        img = get_enhanced_image(datasets.squeeze(), enhance=self.enhancer)
        if 'bands' in img.data.sizes and 'bands' not in datasets.sizes:
            LOG.debug("Datasets without 'bands' become image with 'bands' due to enhancement.")
            LOG.debug("Needs to regenerate mitiff image description")
            image_description = self._make_image_description(img.data, **kwargs)
            tif.SetField(IMAGEDESCRIPTION, (image_description).encode('utf-8'))
        for band in img.data['bands']:
            chn = img.data.sel(bands=band)
            data = chn.values.clip(0, 1) * 254. + 1
            data = data.clip(0, 255)
            tif.write_image(data.astype(np.uint8), compression='deflate')

    def _save_datasets_as_mitiff(self, datasets, image_description,
                                 gen_filename, **kwargs):
        """Put all together and save as a tiff file.

        Include the special tags making it a mitiff file.

        """
        from libtiff import TIFF

        tif = TIFF.open(gen_filename, mode='wb')

        tif.SetField(IMAGEDESCRIPTION, (image_description).encode('utf-8'))

        cns = self.translate_channel_name.get(kwargs['sensor'], {})
        if isinstance(datasets, list):
            LOG.debug("Saving datasets as list")
            for _cn in self.channel_order[kwargs['sensor']]:
                for dataset in datasets:
                    if dataset.attrs['name'] == _cn:
                        # Need to possible translate channels names from satpy to mitiff
                        cn = cns.get(dataset.attrs['name'], dataset.attrs['name'])
                        data = self._calibrate_data(dataset, dataset.attrs['calibration'],
                                                    self.mitiff_config[kwargs['sensor']][cn]['min-val'],
                                                    self.mitiff_config[kwargs['sensor']][cn]['max-val'])
                        tif.write_image(data.astype(np.uint8), compression='deflate')
                        break
        elif 'dataset' in datasets.attrs['name']:
            self._save_single_dataset(datasets, cns, tif, kwargs)
        elif self.palette:
            LOG.debug("Saving dataset as palette.")
            self._save_as_palette(tif, datasets, **kwargs)
        else:
            LOG.debug("Saving datasets as enhanced image")
            self._save_as_enhanced(tif, datasets, **kwargs)
        del tif

    def _save_single_dataset(self, datasets, cns, tif, kwargs):
        LOG.debug("Saving %s as a dataset.", datasets.attrs['name'])
        if len(datasets.dims) == 2 and (all('bands' not in i for i in datasets.dims)):
            # Special case with only one channel ie. no bands

            # Need to possible translate channels names from satpy to mitiff
            # Note the last index is a tuple index.
            cn = cns.get(datasets.attrs['prerequisites'][0]['name'],
                         datasets.attrs['prerequisites'][0]['name'])
            data = self._calibrate_data(datasets, datasets.attrs['prerequisites'][0].get('calibration'),
                                        self.mitiff_config[kwargs['sensor']][cn]['min-val'],
                                        self.mitiff_config[kwargs['sensor']][cn]['max-val'])

            tif.write_image(data.astype(np.uint8), compression='deflate')
        else:
            for _cn_i, _cn in enumerate(self.channel_order[kwargs['sensor']]):
                for band in datasets['bands']:
                    if band == _cn:
                        chn = datasets.sel(bands=band)
                        # Need to possible translate channels names from satpy to mitiff
                        # Note the last index is a tuple index.
                        cn = cns.get(chn.attrs['prerequisites'][_cn_i]['name'],
                                     chn.attrs['prerequisites'][_cn_i]['name'])
                        data = self._calibrate_data(chn, chn.attrs['prerequisites'][_cn_i].get('calibration'),
                                                    self.mitiff_config[kwargs['sensor']][cn]['min-val'],
                                                    self.mitiff_config[kwargs['sensor']][cn]['max-val'])

                        tif.write_image(data.astype(np.uint8), compression='deflate')
                        break



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.nucaps module."""

import datetime
import os
import unittest
from unittest import mock

import numpy as np

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler
from satpy.tests.utils import convert_file_content_to_data_array

DEFAULT_FILE_DTYPE = np.float32
DEFAULT_FILE_SHAPE = (120,)
DEFAULT_PRES_FILE_SHAPE = (120, 100,)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0], dtype=DEFAULT_FILE_DTYPE)
DEFAULT_PRES_FILE_DATA = np.arange(DEFAULT_PRES_FILE_SHAPE[1], dtype=DEFAULT_FILE_DTYPE)
DEFAULT_PRES_FILE_DATA = np.repeat([DEFAULT_PRES_FILE_DATA], DEFAULT_PRES_FILE_SHAPE[0], axis=0)
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[0]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[0]).astype(DEFAULT_FILE_DTYPE)
ALL_PRESSURE_LEVELS = [
    0.0161, 0.0384, 0.0769, 0.137, 0.2244, 0.3454, 0.5064, 0.714,
    0.9753, 1.2972, 1.6872, 2.1526, 2.7009, 3.3398, 4.077, 4.9204,
    5.8776, 6.9567, 8.1655, 9.5119, 11.0038, 12.6492, 14.4559, 16.4318,
    18.5847, 20.9224, 23.4526, 26.1829, 29.121, 32.2744, 35.6505,
    39.2566, 43.1001, 47.1882, 51.5278, 56.126, 60.9895, 66.1253,
    71.5398, 77.2396, 83.231, 89.5204, 96.1138, 103.017, 110.237,
    117.777, 125.646, 133.846, 142.385, 151.266, 160.496, 170.078,
    180.018, 190.32, 200.989, 212.028, 223.441, 235.234, 247.408,
    259.969, 272.919, 286.262, 300, 314.137, 328.675, 343.618, 358.966,
    374.724, 390.893, 407.474, 424.47, 441.882, 459.712, 477.961,
    496.63, 515.72, 535.232, 555.167, 575.525, 596.306, 617.511, 639.14,
    661.192, 683.667, 706.565, 729.886, 753.628, 777.79, 802.371,
    827.371, 852.788, 878.62, 904.866, 931.524, 958.591, 986.067,
    1013.95, 1042.23, 1070.92, 1100
]
ALL_PRESSURE_LEVELS = np.repeat([ALL_PRESSURE_LEVELS], DEFAULT_PRES_FILE_SHAPE[0], axis=0)


class FakeNetCDF4FileHandler2(FakeNetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        file_content = {
            '/attr/time_coverage_start': "2020-10-20T12:00:00.5Z",
            '/attr/time_coverage_end': "2020-10-20T12:00:36Z",
            '/attr/start_orbit_number': 1,
            '/attr/end_orbit_number': 2,
            '/attr/platform_name': 'NPP',
            '/attr/instrument_name': 'CrIS, ATMS, VIIRS',
        }
        for k, units, standard_name in [
            ('Solar_Zenith', 'degrees', 'solar_zenith_angle'),
            ('Topography', 'meters', ''),
            ('Land_Fraction', '1', ''),
            ('Surface_Pressure', 'mb', ''),
            ('Skin_Temperature', 'Kelvin', 'surface_temperature'),
        ]:
            file_content[k] = DEFAULT_FILE_DATA
            file_content[k + '/shape'] = DEFAULT_FILE_SHAPE
            file_content[k + '/attr/units'] = units
            file_content[k + '/attr/valid_range'] = (0., 120.)
            file_content[k + '/attr/_FillValue'] = -9999.
            if standard_name:
                file_content[k + '/attr/standard_name'] = standard_name
        for k, units, standard_name in [
            ('Temperature', 'Kelvin', 'air_temperature'),
            ('Effective_Pressure', 'mb', ''),
            ('H2O', '1', ''),
            ('H2O_MR', 'g/g', ''),
            ('O3', '1', ''),
            ('O3_MR', '1', ''),
            ('Liquid_H2O', '1', ''),
            ('Liquid_H2O_MR', 'g/g', 'cloud_liquid_water_mixing_ratio'),
            ('CO', '1', ''),
            ('CO_MR', '1', ''),
            ('CH4', '1', ''),
            ('CH4_MR', '1', ''),
            ('CO2', '1', ''),
            ('HNO3', '1', ''),
            ('HNO3_MR', '1', ''),
            ('N2O', '1', ''),
            ('N2O_MR', '1', ''),
            ('SO2', '1', ''),
            ('SO2_MR', '1', ''),
        ]:
            file_content[k] = DEFAULT_PRES_FILE_DATA
            file_content[k + '/shape'] = DEFAULT_PRES_FILE_SHAPE
            file_content[k + '/attr/units'] = units
            file_content[k + '/attr/valid_range'] = (0., 120.)
            file_content[k + '/attr/_FillValue'] = -9999.
            if standard_name:
                file_content[k + '/attr/standard_name'] = standard_name
        k = 'Pressure'
        file_content[k] = ALL_PRESSURE_LEVELS
        file_content[k + '/shape'] = DEFAULT_PRES_FILE_SHAPE
        file_content[k + '/attr/units'] = 'mb'
        file_content[k + '/attr/valid_range'] = (0., 2000.)
        file_content[k + '/attr/_FillValue'] = -9999.

        k = 'Quality_Flag'
        file_content[k] = DEFAULT_FILE_DATA.astype(np.int32)
        file_content[k + '/shape'] = DEFAULT_FILE_SHAPE
        file_content[k + '/attr/valid_range'] = (0, 31)
        file_content[k + '/attr/_FillValue'] = -9999.

        k = 'Longitude'
        file_content[k] = DEFAULT_LON_DATA
        file_content[k + '/shape'] = DEFAULT_FILE_SHAPE
        file_content[k + '/attr/units'] = 'degrees_east'
        file_content[k + '/attr/valid_range'] = (-180., 180.)
        file_content[k + '/attr/standard_name'] = 'longitude'
        file_content[k + '/attr/_FillValue'] = -9999.

        k = 'Latitude'
        file_content[k] = DEFAULT_LAT_DATA
        file_content[k + '/shape'] = DEFAULT_FILE_SHAPE
        file_content[k + '/attr/units'] = 'degrees_north'
        file_content[k + '/attr/valid_range'] = (-90., 90.)
        file_content[k + '/attr/standard_name'] = 'latitude'
        file_content[k + '/attr/_FillValue'] = -9999.

        attrs = ('_FillValue', 'flag_meanings', 'flag_values', 'units')
        cris_fors_dim_name = 'Number_of_CrIS_FORs'
        pressure_levels_dim_name = 'Number_of_P_Levels'
        if ('_v1' in filename):
            cris_fors_dim_name = 'number_of_FORs'
            pressure_levels_dim_name = 'number_of_p_levels'
        convert_file_content_to_data_array(
            file_content, attrs=attrs,
            dims=('z', cris_fors_dim_name, pressure_levels_dim_name))
        return file_content


class TestNUCAPSReader(unittest.TestCase):
    """Test NUCAPS Reader."""

    yaml_file = "nucaps.yaml"

    def setUp(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.nucaps import NUCAPSFileHandler
        self.reader_configs = config_search_paths(os.path.join('readers', self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(NUCAPSFileHandler, '__bases__', (FakeNetCDF4FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
        ])
        self.assertEqual(len(loadables), 1)
        r.create_filehandlers(loadables)
        # make sure we have some files
        self.assertTrue(r.file_handlers)

    def test_init_with_kwargs(self):
        """Test basic init with extra parameters."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs, mask_surface=False)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
        ])
        self.assertEqual(len(loadables), 1)
        r.create_filehandlers(loadables, fh_kwargs={'mask_surface': False})
        # make sure we have some files
        self.assertTrue(r.file_handlers)

    def test_load_nonpressure_based(self):
        """Test loading all channels that aren't based on pressure."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Solar_Zenith',
                           'Topography',
                           'Land_Fraction',
                           'Surface_Pressure',
                           'Skin_Temperature',
                           'Quality_Flag',
                           ])
        self.assertEqual(len(datasets), 6)
        for v in datasets.values():
            # self.assertNotEqual(v.info['resolution'], 0)
            # self.assertEqual(v.info['units'], 'degrees')
            self.assertEqual(v.ndim, 1)
            self.assertEqual(v.attrs['sensor'], set(['cris', 'atms', 'viirs']))
            self.assertEqual(type(v.attrs['start_time']), datetime.datetime)
            self.assertEqual(type(v.attrs['end_time']), datetime.datetime)

    def test_load_pressure_based(self):
        """Test loading all channels based on pressure."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Temperature',
                           'Effective_Pressure',
                           'H2O',
                           'H2O_MR',
                           'O3',
                           'O3_MR',
                           'Liquid_H2O',
                           'Liquid_H2O_MR',
                           'CO',
                           'CO_MR',
                           'CH4',
                           'CH4_MR',
                           'CO2',
                           'HNO3',
                           'HNO3_MR',
                           'N2O',
                           'N2O_MR',
                           'SO2',
                           'SO2_MR',
                           ])
        self.assertEqual(len(datasets), 19)
        for v in datasets.values():
            # self.assertNotEqual(v.info['resolution'], 0)
            self.assertEqual(v.ndim, 2)
            if np.issubdtype(v.dtype, np.floating):
                assert '_FillValue' not in v.attrs

    def test_load_multiple_files_pressure(self):
        """Test loading Temperature from multiple input files."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
            'NUCAPS-EDR_v1r0_npp_s201603011159009_e201603011159307_c201603011222270.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names['Temperature'], pressure_levels=True)
        self.assertEqual(len(datasets), 100)
        for v in datasets.values():
            self.assertEqual(v.ndim, 1)

    def test_load_individual_pressure_levels_true(self):
        """Test loading Temperature with individual pressure datasets."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names['Temperature'], pressure_levels=True)
        self.assertEqual(len(datasets), 100)
        for v in datasets.values():
            self.assertEqual(v.ndim, 1)

    def test_load_individual_pressure_levels_min_max(self):
        """Test loading individual Temperature with min/max level specified."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names['Temperature'], pressure_levels=(100., 150.))
        self.assertEqual(len(datasets), 6)
        for v in datasets.values():
            self.assertEqual(v.ndim, 1)

    def test_load_individual_pressure_levels_single(self):
        """Test loading individual Temperature with specific levels."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names['Temperature'], pressure_levels=(103.017,))
        self.assertEqual(len(datasets), 1)
        for v in datasets.values():
            self.assertEqual(v.ndim, 1)

    def test_load_pressure_levels_true(self):
        """Test loading Temperature with all pressure levels."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Temperature'], pressure_levels=True)
        self.assertEqual(len(datasets), 1)
        for v in datasets.values():
            self.assertEqual(v.ndim, 2)
            self.assertTupleEqual(v.shape, DEFAULT_PRES_FILE_SHAPE)

    def test_load_pressure_levels_min_max(self):
        """Test loading Temperature with min/max level specified."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Temperature'], pressure_levels=(100., 150.))
        self.assertEqual(len(datasets), 1)
        for v in datasets.values():
            self.assertEqual(v.ndim, 2)
            self.assertTupleEqual(v.shape,
                                  (DEFAULT_PRES_FILE_SHAPE[0], 6))

    def test_load_pressure_levels_single(self):
        """Test loading a specific Temperature level."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Temperature'], pressure_levels=(103.017,))
        self.assertEqual(len(datasets), 1)
        for v in datasets.values():
            self.assertEqual(v.ndim, 2)
            self.assertTupleEqual(v.shape,
                                  (DEFAULT_PRES_FILE_SHAPE[0], 1))

    def test_load_pressure_levels_single_and_pressure_levels(self):
        """Test loading a specific Temperature level and pressure levels."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-EDR_v1r0_npp_s201603011158009_e201603011158307_c201603011222270.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Temperature', 'Pressure_Levels'], pressure_levels=(103.017,))
        self.assertEqual(len(datasets), 2)
        t_ds = datasets['Temperature']
        self.assertEqual(t_ds.ndim, 2)
        self.assertTupleEqual(t_ds.shape,
                              (DEFAULT_PRES_FILE_SHAPE[0], 1))
        pl_ds = datasets['Pressure_Levels']
        self.assertTupleEqual(pl_ds.shape, (1,))


class TestNUCAPSScienceEDRReader(unittest.TestCase):
    """Test NUCAPS Science EDR Reader."""

    yaml_file = "nucaps.yaml"

    def setUp(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.nucaps import NUCAPSFileHandler
        self.reader_configs = config_search_paths(os.path.join('readers', self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(NUCAPSFileHandler, '__bases__', (FakeNetCDF4FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc',
        ])
        self.assertEqual(len(loadables), 1)
        r.create_filehandlers(loadables)
        # make sure we have some files
        self.assertTrue(r.file_handlers)

    def test_load_nonpressure_based(self):
        """Test loading all channels that aren't based on pressure."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Topography',
                           'Land_Fraction',
                           'Surface_Pressure',
                           'Skin_Temperature',
                           'Quality_Flag',
                           ])
        self.assertEqual(len(datasets), 5)
        for v in datasets.values():
            self.assertEqual(v.ndim, 1)
            self.assertEqual(v.attrs['sensor'], set(['cris', 'atms', 'viirs']))
            self.assertEqual(type(v.attrs['start_time']), datetime.datetime)
            self.assertEqual(type(v.attrs['end_time']), datetime.datetime)

    def test_load_pressure_based(self):
        """Test loading all channels based on pressure."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Temperature',
                           'H2O',
                           'H2O_MR',
                           'O3',
                           'O3_MR',
                           'CO',
                           'CO_MR',
                           'CH4',
                           'CH4_MR',
                           'CO2',
                           'HNO3',
                           'HNO3_MR',
                           'N2O',
                           'N2O_MR',
                           'SO2',
                           'SO2_MR',
                           ])
        self.assertEqual(len(datasets), 16)
        for v in datasets.values():
            # self.assertNotEqual(v.info['resolution'], 0)
            self.assertEqual(v.ndim, 2)

    def test_load_individual_pressure_levels_true(self):
        """Test loading Temperature with individual pressure datasets."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names['Temperature'], pressure_levels=True)
        self.assertEqual(len(datasets), 100)
        for v in datasets.values():
            self.assertEqual(v.ndim, 1)

    def test_load_individual_pressure_levels_min_max(self):
        """Test loading individual Temperature with min/max level specified."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names['Temperature'], pressure_levels=(100., 150.))
        self.assertEqual(len(datasets), 6)
        for v in datasets.values():
            self.assertEqual(v.ndim, 1)

    def test_load_individual_pressure_levels_single(self):
        """Test loading individual Temperature with specific levels."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(r.pressure_dataset_names['Temperature'], pressure_levels=(103.017,))
        self.assertEqual(len(datasets), 1)
        for v in datasets.values():
            self.assertEqual(v.ndim, 1)

    def test_load_pressure_levels_true(self):
        """Test loading Temperature with all pressure levels."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Temperature'], pressure_levels=True)
        self.assertEqual(len(datasets), 1)
        for v in datasets.values():
            self.assertEqual(v.ndim, 2)
            self.assertTupleEqual(v.shape, DEFAULT_PRES_FILE_SHAPE)

    def test_load_pressure_levels_min_max(self):
        """Test loading Temperature with min/max level specified."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Temperature'], pressure_levels=(100., 150.))
        self.assertEqual(len(datasets), 1)
        for v in datasets.values():
            self.assertEqual(v.ndim, 2)
            self.assertTupleEqual(v.shape,
                                  (DEFAULT_PRES_FILE_SHAPE[0], 6))

    def test_load_pressure_levels_single(self):
        """Test loading a specific Temperature level."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Temperature'], pressure_levels=(103.017,))
        self.assertEqual(len(datasets), 1)
        for v in datasets.values():
            self.assertEqual(v.ndim, 2)
            self.assertTupleEqual(v.shape,
                                  (DEFAULT_PRES_FILE_SHAPE[0], 1))

    def test_load_pressure_levels_single_and_pressure_levels(self):
        """Test loading a specific Temperature level and pressure levels."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'NUCAPS-sciEDR_am_npp_s20190703223319_e20190703223349_STC_fsr.nc',
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['Temperature', 'Pressure_Levels'], pressure_levels=(103.017,))
        self.assertEqual(len(datasets), 2)
        t_ds = datasets['Temperature']
        self.assertEqual(t_ds.ndim, 2)
        self.assertTupleEqual(t_ds.shape,
                              (DEFAULT_PRES_FILE_SHAPE[0], 1))
        pl_ds = datasets['Pressure_Levels']
        self.assertTupleEqual(pl_ds.shape, (1,))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module defining various utilities."""

from __future__ import annotations

import contextlib
import logging
import os
import warnings
from typing import Mapping

import numpy as np
import xarray as xr
import yaml
from yaml import BaseLoader

try:
    from yaml import UnsafeLoader
except ImportError:
    from yaml import Loader as UnsafeLoader  # type: ignore

_is_logging_on = False
TRACE_LEVEL = 5


def ensure_dir(filename):
    """Check if the dir of f exists, otherwise create it."""
    directory = os.path.dirname(filename)
    if directory and not os.path.isdir(directory):
        os.makedirs(directory)


def debug_on(deprecation_warnings=True):
    """Turn debugging logging on.

    Sets up a StreamHandler to to `sys.stderr` at debug level for all
    loggers, such that all debug messages (and log messages with higher
    severity) are logged to the standard error stream.

    By default, since Satpy 0.26, this also enables the global visibility
    of deprecation warnings.  This can be suppressed by passing a false
    value.

    Args:
        deprecation_warnings (Optional[bool]): Switch on deprecation warnings.
            Defaults to True.

    Returns:
        None
    """
    logging_on(logging.DEBUG)
    if deprecation_warnings:
        deprecation_warnings_on()


def debug_off():
    """Turn debugging logging off.

    This disables both debugging logging and the global visibility of
    deprecation warnings.
    """
    logging_off()
    deprecation_warnings_off()


@contextlib.contextmanager
def debug(deprecation_warnings=True):
    """Context manager to temporarily set debugging on.

    Example::

        >>> with satpy.utils.debug():
        ...     code_here()

    Args:
        deprecation_warnings (Optional[bool]): Switch on deprecation warnings.
            Defaults to True.
    """
    debug_on(deprecation_warnings=deprecation_warnings)
    yield
    debug_off()


def trace_on():
    """Turn trace logging on."""
    logging_on(TRACE_LEVEL)


class _WarningManager:
    """Class to handle switching warnings on and off."""

    filt = None


_warning_manager = _WarningManager()


def deprecation_warnings_on():
    """Switch on deprecation warnings."""
    warnings.filterwarnings("default", category=DeprecationWarning)
    _warning_manager.filt = warnings.filters[0]


def deprecation_warnings_off():
    """Switch off deprecation warnings."""
    if _warning_manager.filt in warnings.filters:
        warnings.filters.remove(_warning_manager.filt)


def logging_on(level=logging.WARNING):
    """Turn logging on."""
    global _is_logging_on

    if not _is_logging_on:
        console = logging.StreamHandler()
        console.setFormatter(logging.Formatter("[%(levelname)s: %(asctime)s :"
                                               " %(name)s] %(message)s",
                                               '%Y-%m-%d %H:%M:%S'))
        console.setLevel(level)
        logging.getLogger('').addHandler(console)
        _is_logging_on = True

    log = logging.getLogger('')
    log.setLevel(level)
    for h in log.handlers:
        h.setLevel(level)


def logging_off():
    """Turn logging off."""
    logging.getLogger('').handlers = [logging.NullHandler()]


def get_logger(name):
    """Return logger with null handler added if needed."""
    if not hasattr(logging.Logger, 'trace'):
        logging.addLevelName(TRACE_LEVEL, 'TRACE')

        def trace(self, message, *args, **kwargs):
            if self.isEnabledFor(TRACE_LEVEL):
                # Yes, logger takes its '*args' as 'args'.
                self._log(TRACE_LEVEL, message, args, **kwargs)

        logging.Logger.trace = trace

    log = logging.getLogger(name)
    return log


def in_ipynb():
    """Check if we are in a jupyter notebook."""
    try:
        return 'ZMQ' in get_ipython().__class__.__name__
    except NameError:
        return False


# Spherical conversions


def lonlat2xyz(lon, lat):
    """Convert lon lat to cartesian."""
    lat = np.deg2rad(lat)
    lon = np.deg2rad(lon)
    x = np.cos(lat) * np.cos(lon)
    y = np.cos(lat) * np.sin(lon)
    z = np.sin(lat)
    return x, y, z


def xyz2lonlat(x, y, z, asin=False):
    """Convert cartesian to lon lat."""
    lon = np.rad2deg(np.arctan2(y, x))
    if asin:
        lat = np.rad2deg(np.arcsin(z))
    else:
        lat = np.rad2deg(np.arctan2(z, np.sqrt(x ** 2 + y ** 2)))
    return lon, lat


def angle2xyz(azi, zen):
    """Convert azimuth and zenith to cartesian."""
    azi = np.deg2rad(azi)
    zen = np.deg2rad(zen)
    x = np.sin(zen) * np.sin(azi)
    y = np.sin(zen) * np.cos(azi)
    z = np.cos(zen)
    return x, y, z


def xyz2angle(x, y, z, acos=False):
    """Convert cartesian to azimuth and zenith."""
    azi = np.rad2deg(np.arctan2(x, y))
    if acos:
        zen = np.rad2deg(np.arccos(z))
    else:
        zen = 90 - np.rad2deg(np.arctan2(z, np.sqrt(x ** 2 + y ** 2)))
    return azi, zen


def proj_units_to_meters(proj_str):
    """Convert projection units from kilometers to meters."""
    proj_parts = proj_str.split()
    new_parts = []
    for itm in proj_parts:
        key, val = itm.split('=')
        key = key.strip('+')
        if key in ['a', 'b', 'h']:
            val = float(val)
            if val < 6e6:
                val *= 1000.
                val = '%.3f' % val

        if key == 'units' and val == 'km':
            continue

        new_parts.append('+%s=%s' % (key, val))

    return ' '.join(new_parts)


def _get_sunz_corr_li_and_shibata(cos_zen):
    return 24.35 / (2. * cos_zen + np.sqrt(498.5225 * cos_zen**2 + 1))


def atmospheric_path_length_correction(data, cos_zen, limit=88., max_sza=95.):
    """Perform Sun zenith angle correction.

    This function uses the correction method proposed by
    Li and Shibata (2006): https://doi.org/10.1175/JAS3682.1

    The correction is limited to ``limit`` degrees (default: 88.0 degrees). For
    larger zenith angles, the correction is the same as at the ``limit`` if
    ``max_sza`` is `None`. The default behavior is to gradually reduce the
    correction past ``limit`` degrees up to ``max_sza`` where the correction
    becomes 0. Both ``data`` and ``cos_zen`` should be 2D arrays of the same
    shape.

    """
    # Convert the zenith angle limit to cosine of zenith angle
    limit_rad = np.deg2rad(limit)
    limit_cos = np.cos(limit_rad)
    max_sza_rad = np.deg2rad(max_sza) if max_sza is not None else max_sza

    # Cosine correction
    corr = _get_sunz_corr_li_and_shibata(cos_zen)
    # Use constant value (the limit) for larger zenith angles
    corr_lim = _get_sunz_corr_li_and_shibata(limit_cos)

    if max_sza is not None:
        # gradually fall off for larger zenith angle
        grad_factor = (np.arccos(cos_zen) - limit_rad) / (max_sza_rad - limit_rad)
        # invert the factor so maximum correction is done at `limit` and falls off later
        grad_factor = 1. - np.log(grad_factor + 1) / np.log(2)
        # make sure we don't make anything negative
        grad_factor = grad_factor.clip(0.)
    else:
        # Use constant value (the limit) for larger zenith angles
        grad_factor = 1.
    corr = corr.where(cos_zen > limit_cos, grad_factor * corr_lim)
    # Force "night" pixels to 0 (where SZA is invalid)
    corr = corr.where(cos_zen.notnull(), 0)

    return data * corr


def get_satpos(dataset):
    """Get satellite position from dataset attributes.

    Preferences are:

    * Longitude & Latitude: Nadir, actual, nominal, projection
    * Altitude: Actual, nominal, projection

    A warning is issued when projection values have to be used because nothing else is available.

    Returns:
        Geodetic longitude, latitude, altitude

    """
    try:
        orb_params = dataset.attrs['orbital_parameters']

        alt = _get_sat_altitude(orb_params)

        lon, lat = _get_sat_lonlat(orb_params)
    except KeyError:
        # Legacy
        lon = dataset.attrs['satellite_longitude']
        lat = dataset.attrs['satellite_latitude']
        alt = dataset.attrs['satellite_altitude']

    return lon, lat, alt


def _get_sat_altitude(orb_params):
    # Altitude
    try:
        alt = orb_params['satellite_actual_altitude']
    except KeyError:
        try:
            alt = orb_params['satellite_nominal_altitude']
        except KeyError:
            alt = orb_params['projection_altitude']
            warnings.warn('Actual satellite altitude not available, using projection altitude instead.')
    return alt


def _get_sat_lonlat(orb_params):
    # Longitude & Latitude
    try:
        lon = orb_params['nadir_longitude']
        lat = orb_params['nadir_latitude']
    except KeyError:
        try:
            lon = orb_params['satellite_actual_longitude']
            lat = orb_params['satellite_actual_latitude']
        except KeyError:
            try:
                lon = orb_params['satellite_nominal_longitude']
                lat = orb_params['satellite_nominal_latitude']
            except KeyError:
                lon = orb_params['projection_longitude']
                lat = orb_params['projection_latitude']
                warnings.warn('Actual satellite lon/lat not available, using projection centre instead.')
    return lon, lat


def recursive_dict_update(d, u):
    """Recursive dictionary update.

    Copied from:

        http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth

    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = recursive_dict_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def _check_yaml_configs(configs, key):
    """Get a diagnostic for the yaml *configs*.

    *key* is the section to look for to get a name for the config at hand.
    """
    diagnostic = {}
    for i in configs:
        for fname in i:
            with open(fname, 'r', encoding='utf-8') as stream:
                try:
                    res = yaml.load(stream, Loader=UnsafeLoader)
                    msg = 'ok'
                except yaml.YAMLError as err:
                    stream.seek(0)
                    res = yaml.load(stream, Loader=BaseLoader)
                    if err.context == 'while constructing a Python object':
                        msg = err.problem
                    else:
                        msg = 'error'
                finally:
                    try:
                        diagnostic[res[key]['name']] = msg
                    except (KeyError, TypeError):
                        # this object doesn't have a 'name'
                        pass
    return diagnostic


def _check_import(module_names):
    """Import the specified modules and provide status."""
    diagnostics = {}
    for module_name in module_names:
        try:
            __import__(module_name)
            res = 'ok'
        except ImportError as err:
            res = str(err)
        diagnostics[module_name] = res
    return diagnostics


def check_satpy(readers=None, writers=None, extras=None):
    """Check the satpy readers and writers for correct installation.

    Args:
        readers (list or None): Limit readers checked to those specified
        writers (list or None): Limit writers checked to those specified
        extras (list or None): Limit extras checked to those specified

    Returns: bool
        True if all specified features were successfully loaded.

    """
    from satpy.readers import configs_for_reader
    from satpy.writers import configs_for_writer

    print('Readers')
    print('=======')
    for reader, res in sorted(_check_yaml_configs(configs_for_reader(reader=readers), 'reader').items()):
        print(reader + ': ', res)
    print()

    print('Writers')
    print('=======')
    for writer, res in sorted(_check_yaml_configs(configs_for_writer(writer=writers), 'writer').items()):
        print(writer + ': ', res)
    print()

    print('Extras')
    print('======')
    module_names = extras if extras is not None else ('cartopy', 'geoviews')
    for module_name, res in sorted(_check_import(module_names).items()):
        print(module_name + ': ', res)
    print()


def unify_chunks(*data_arrays: xr.DataArray) -> tuple[xr.DataArray, ...]:
    """Run :func:`xarray.unify_chunks` if input dimensions are all the same size.

    This is mostly used in :class:`satpy.composites.CompositeBase` to safe
    guard against running :func:`dask.array.core.map_blocks` with arrays of
    different chunk sizes. Doing so can cause unexpected results or errors.
    However, xarray's ``unify_chunks`` will raise an exception if dimensions
    of the provided DataArrays are different sizes. This is a common case for
    Satpy. For example, the "bands" dimension may be 1 (L), 2 (LA), 3 (RGB), or
    4 (RGBA) for most compositor operations that combine other composites
    together.

    """
    if not hasattr(xr, "unify_chunks"):
        return data_arrays
    if not _all_dims_same_size(data_arrays):
        return data_arrays
    return tuple(xr.unify_chunks(*data_arrays))


def _all_dims_same_size(data_arrays: tuple[xr.DataArray, ...]) -> bool:
    known_sizes: dict[str, int] = {}
    for data_arr in data_arrays:
        for dim, dim_size in data_arr.sizes.items():
            known_size = known_sizes.setdefault(dim, dim_size)
            if dim_size != known_size:
                # this dimension is a different size than previously found
                # xarray.unify_chunks will error out if we tried to use it
                return False
    return True


@contextlib.contextmanager
def ignore_invalid_float_warnings():
    """Ignore warnings generated for working with NaN/inf values.

    Numpy and dask sometimes don't like NaN or inf values in normal function
    calls. This context manager hides/ignores them inside its context.

    Examples:
        Use around numpy operations that you expect to produce warnings::

            with ignore_invalid_float_warnings():
                np.nanmean(np.nan)

    """
    with np.errstate(invalid="ignore"), warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)
        yield



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Simple command line too that converts an image into a NinJo Tiff file.

NinJo Tiff metadata can be passed as command line input or
through a config file (an example is given in the ninjo-cmd.yaml
file in the etc directory).

The area matching the input image shall be defined in the
areas configuration file (located in $PPP_CONFIG_DIR).

"""

import argparse
import os

import yaml

from satpy import Scene
from satpy.pyresample import get_area_def
from satpy.utils import debug_on

try:
    from yaml import UnsafeLoader
except ImportError:
    from yaml import Loader as UnsafeLoader  # type: ignore


debug_on()

parser = argparse.ArgumentParser(description='Turn an image into a NinjoTiff.')
parser.add_argument('--cfg', dest='cfg', action="store",
                    help="YAML configuration as an alternative to the command line input for NinJo metadata.")
parser.add_argument('--input_dir', dest='input_dir', action="store",
                    help="Directory with input data, that must contain a timestamp in the filename.")
parser.add_argument('--chan_id', dest='chan_id', action="store", help="Channel ID", default="9999")
parser.add_argument('--sat_id', dest='sat_id', action="store", help="Satellite ID", default="8888")
parser.add_argument('--data_cat', dest='data_cat', action="store",
                    help="Category of data (one of GORN, GPRN, PORN)", default="GORN")
parser.add_argument('--area', dest='areadef', action="store",
                    help="Area name, the definition must exist in your areas configuration file",
                    default="nrEURO1km_NPOL_COALeqc")
parser.add_argument('--ph_unit', dest='ph_unit', action="store", help="Physical unit", default="CELSIUS")
parser.add_argument('--data_src', dest='data_src', action="store", help="Data source", default="EUMETCAST")
args = parser.parse_args()

if (args.input_dir is not None):
    os.chdir(args.input_dir)

cfg = vars(args)
if (args.cfg is not None):
    with open(args.cfg, 'r') as ymlfile:
        cfg = yaml.load(ymlfile, Loader=UnsafeLoader)

narea = get_area_def(args.areadef)
global_data = Scene(reader="generic_image")
global_data.load(['image'])

global_data['image'].info['area'] = narea
fname = global_data['image'].info['filename']
ofname = fname[:-3] + "tif"

# global_data.save_dataset('image', filename="out.png", writer="simple_image")
global_data.save_dataset('image', filename=ofname, writer="ninjotiff",
                         sat_id=cfg['sat_id'],
                         chan_id=cfg['chan_id'],
                         data_cat=cfg['data_cat'],
                         data_source=cfg['data_src'],
                         physic_unit=cfg['ph_unit'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""SAFE SAR-C reader.

This module implements a reader for Sentinel 1 SAR-C GRD (level1) SAFE format as
provided by ESA. The format is comprised of a directory containing multiple
files, most notably two measurement files in geotiff and a few xml files for
calibration, noise and metadata.

References:
  - *Level 1 Product Formatting*
    https://sentinel.esa.int/web/sentinel/technical-guides/sentinel-1-sar/products-algorithms/level-1-product-formatting

  - J. Park, A. A. Korosov, M. Babiker, S. Sandven and J. Won,
    *"Efficient Thermal Noise Removal for Sentinel-1 TOPSAR Cross-Polarization Channel,"*
    in IEEE Transactions on Geoscience and Remote Sensing, vol. 56, no. 3,
    pp. 1555-1565, March 2018.
    doi: `10.1109/TGRS.2017.2765248 <https://doi.org/10.1109/TGRS.2017.2765248>`_

"""

import logging
from functools import lru_cache
from threading import Lock

import defusedxml.ElementTree as ET
import numpy as np
import rasterio
import rioxarray
import xarray as xr
from dask import array as da
from dask.base import tokenize
from xarray import DataArray

from satpy import CHUNK_SIZE
from satpy.readers.file_handlers import BaseFileHandler

logger = logging.getLogger(__name__)


def dictify(r):
    """Convert an ElementTree into a dict."""
    return {r.tag: _dictify(r)}


def _dictify(r):
    """Convert an xml element to dict."""
    d = {}
    if r.text and r.text.strip():
        try:
            return int(r.text)
        except ValueError:
            try:
                return float(r.text)
            except ValueError:
                return r.text
    for x in r.findall("./*"):
        if x.tag in d and not isinstance(d[x.tag], list):
            d[x.tag] = [d[x.tag]]
            d[x.tag].append(_dictify(x))
        else:
            d[x.tag] = _dictify(x)
    return d


def _get_calibration_name(calibration):
    """Get the proper calibration name."""
    calibration_name = getattr(calibration, "name", calibration) or 'gamma'
    if calibration_name == 'sigma_nought':
        calibration_name = 'sigmaNought'
    elif calibration_name == 'beta_nought':
        calibration_name = 'betaNought'
    return calibration_name


class SAFEXML(BaseFileHandler):
    """XML file reader for the SAFE format."""

    def __init__(self, filename, filename_info, filetype_info,
                 header_file=None):
        """Init the xml filehandler."""
        super(SAFEXML, self).__init__(filename, filename_info, filetype_info)

        self._start_time = filename_info['start_time']
        self._end_time = filename_info['end_time']
        self._polarization = filename_info['polarization']
        self.root = ET.parse(self.filename)
        self.hdr = {}
        if header_file is not None:
            self.hdr = header_file.get_metadata()
        else:
            self.hdr = self.get_metadata()
        self._image_shape = (self.hdr['product']['imageAnnotation']['imageInformation']['numberOfLines'],
                             self.hdr['product']['imageAnnotation']['imageInformation']['numberOfSamples'])

    def get_metadata(self):
        """Convert the xml metadata to dict."""
        return dictify(self.root.getroot())

    @property
    def start_time(self):
        """Get the start time."""
        return self._start_time

    @property
    def end_time(self):
        """Get the end time."""
        return self._end_time


class SAFEXMLAnnotation(SAFEXML):
    """XML file reader for the SAFE format, Annotation file."""

    def get_dataset(self, key, info, chunks=None):
        """Load a dataset."""
        if self._polarization != key["polarization"]:
            return

        if key["name"] == "incidence_angle":
            return self.get_incidence_angle(chunks=chunks or CHUNK_SIZE)

    @lru_cache(maxsize=10)
    def get_incidence_angle(self, chunks):
        """Get the incidence angle array."""
        incidence_angle = XMLArray(self.root, ".//geolocationGridPoint", "incidenceAngle")
        return incidence_angle.expand(self._image_shape, chunks=chunks)


class SAFEXMLCalibration(SAFEXML):
    """XML file reader for the SAFE format, Calibration file."""

    def get_dataset(self, key, info, chunks=None):
        """Load a dataset."""
        if self._polarization != key["polarization"]:
            return
        if key["name"] == "calibration_constant":
            return self.get_calibration_constant()
        return self.get_calibration(key["name"], chunks=chunks or CHUNK_SIZE)

    def get_calibration_constant(self):
        """Load the calibration constant."""
        return float(self.root.find('.//absoluteCalibrationConstant').text)

    @lru_cache(maxsize=10)
    def get_calibration(self, calibration, chunks=None):
        """Get the calibration array."""
        calibration_name = _get_calibration_name(calibration)
        calibration_vector = self._get_calibration_vector(calibration_name, chunks)
        return calibration_vector

    def _get_calibration_vector(self, calibration_name, chunks):
        """Get the calibration vector."""
        calibration_vector = XMLArray(self.root, ".//calibrationVector", calibration_name)
        return calibration_vector.expand(self._image_shape, chunks=chunks)


class SAFEXMLNoise(SAFEXML):
    """XML file reader for the SAFE format, Noise file."""

    def __init__(self, filename, filename_info, filetype_info,
                 header_file=None):
        """Init the xml filehandler."""
        super().__init__(filename, filename_info, filetype_info, header_file)

        self.azimuth_noise_reader = AzimuthNoiseReader(self.root, self._image_shape)

    def get_dataset(self, key, info, chunks=None):
        """Load a dataset."""
        if self._polarization != key["polarization"]:
            return
        if key["name"] == "noise":
            return self.get_noise_correction(chunks=chunks or CHUNK_SIZE)

    @lru_cache(maxsize=10)
    def get_noise_correction(self, chunks=None):
        """Get the noise correction array."""
        try:
            noise = self.read_legacy_noise(chunks)
        except KeyError:
            range_noise = self.read_range_noise_array(chunks)
            azimuth_noise = self.azimuth_noise_reader.read_azimuth_noise_array(chunks)
            noise = range_noise * azimuth_noise
        return noise

    def read_legacy_noise(self, chunks):
        """Read noise for legacy GRD data."""
        noise = XMLArray(self.root, ".//noiseVector", "noiseLut")
        return noise.expand(self._image_shape, chunks)

    def read_range_noise_array(self, chunks):
        """Read the range-noise array."""
        range_noise = XMLArray(self.root, ".//noiseRangeVector", "noiseRangeLut")
        return range_noise.expand(self._image_shape, chunks)


class AzimuthNoiseReader:
    """Class to parse and read azimuth-noise data.

    The azimuth noise vector is provided as a series of blocks, each comprised
    of a column of data to fill the block and a start and finish column number,
    and a start and finish line.
    For example, we can see here a (fake) azimuth noise array::

        [[ 1.  1.  1. nan nan nan nan nan nan nan]
         [ 1.  1.  1. nan nan nan nan nan nan nan]
         [ 2.  2.  3.  3.  3.  4.  4.  4.  4. nan]
         [ 2.  2.  3.  3.  3.  4.  4.  4.  4. nan]
         [ 2.  2.  3.  3.  3.  4.  4.  4.  4. nan]
         [ 2.  2.  5.  5.  5.  5.  6.  6.  6.  6.]
         [ 2.  2.  5.  5.  5.  5.  6.  6.  6.  6.]
         [ 2.  2.  5.  5.  5.  5.  6.  6.  6.  6.]
         [ 2.  2.  7.  7.  7.  7.  7.  8.  8.  8.]
         [ 2.  2.  7.  7.  7.  7.  7.  8.  8.  8.]]

    As is shown here, the blocks may not cover the full array, and hence it has
    to be gap-filled with NaNs.
    """

    def __init__(self, root, shape):
        """Set up the azimuth noise reader."""
        self.root = root
        self.elements = self.root.findall(".//noiseAzimuthVector")
        self._image_shape = shape
        self.blocks = []

    def read_azimuth_noise_array(self, chunks=CHUNK_SIZE):
        """Read the azimuth noise vectors."""
        self._read_azimuth_noise_blocks(chunks)
        populated_array = self._assemble_azimuth_noise_blocks(chunks)

        return populated_array

    def _read_azimuth_noise_blocks(self, chunks):
        """Read the azimuth noise blocks."""
        self.blocks = []
        for elt in self.elements:
            block = _AzimuthBlock(elt)
            new_arr = block.expand(chunks)
            self.blocks.append(new_arr)

    def _assemble_azimuth_noise_blocks(self, chunks):
        """Assemble the azimuth noise blocks into one single array."""
        # The strategy here is a bit convoluted. The job would be trivial if
        # performed on regular numpy arrays, but here we want to keep the data
        # as xarray/dask array as much as possible.
        # Using a pure xarray approach was tested (with `combine_first`,
        # `interpolate_na`, etc), but was found to be memory-hungry at the time
        # of implementation (March 2021). Hence the usage of a custom algorithm,
        # relying mostly on dask arrays.
        slices = self._create_dask_slices_from_blocks(chunks)
        populated_array = da.vstack(slices).rechunk(chunks)
        populated_array = xr.DataArray(populated_array, dims=['y', 'x'],
                                       coords={'x': np.arange(self._image_shape[1]),
                                               'y': np.arange(self._image_shape[0])})
        return populated_array

    def _create_dask_slices_from_blocks(self, chunks):
        """Create full-width slices from azimuth noise blocks."""
        current_line = 0
        slices = []
        while current_line < self._image_shape[0]:
            new_slice = self._create_dask_slice_from_block_line(current_line, chunks)
            slices.append(new_slice)
            current_line += new_slice.shape[0]
        return slices

    def _create_dask_slice_from_block_line(self, current_line, chunks):
        """Create a dask slice from the blocks at the current line."""
        pieces = self._get_array_pieces_for_current_line(current_line)
        dask_pieces = self._get_padded_dask_pieces(pieces, chunks)
        new_slice = da.hstack(dask_pieces)

        return new_slice

    def _get_array_pieces_for_current_line(self, current_line):
        """Get the array pieces that cover the current line."""
        current_blocks = self._find_blocks_covering_line(current_line)
        current_blocks.sort(key=(lambda x: x.coords['x'][0]))
        next_line = self._get_next_start_line(current_blocks, current_line)
        current_y = np.arange(current_line, next_line)
        pieces = [arr.sel(y=current_y) for arr in current_blocks]
        return pieces

    def _find_blocks_covering_line(self, current_line):
        """Find the blocks covering a given line."""
        current_blocks = []
        for block in self.blocks:
            if block.coords['y'][0] <= current_line <= block.coords['y'][-1]:
                current_blocks.append(block)
        return current_blocks

    def _get_next_start_line(self, current_blocks, current_line):
        next_line = min((arr.coords['y'][-1] for arr in current_blocks)) + 1
        blocks_starting_soon = [block for block in self.blocks if current_line < block.coords["y"][0] < next_line]
        if blocks_starting_soon:
            next_start_line = min((arr.coords["y"][0] for arr in blocks_starting_soon))
            next_line = min(next_line, next_start_line)
        return next_line

    def _get_padded_dask_pieces(self, pieces, chunks):
        """Get the padded pieces of a slice."""
        pieces = sorted(pieces, key=(lambda x: x.coords['x'][0]))
        dask_pieces = []
        previous_x_end = -1
        piece = pieces[0]
        next_x_start = piece.coords['x'][0].item()
        y_shape = len(piece.coords['y'])

        x_shape = (next_x_start - previous_x_end - 1)
        self._fill_dask_pieces(dask_pieces, (y_shape, x_shape), chunks)

        for i, piece in enumerate(pieces):
            dask_pieces.append(piece.data)
            previous_x_end = piece.coords['x'][-1].item()
            try:
                next_x_start = pieces[i + 1].coords['x'][0].item()
            except IndexError:
                next_x_start = self._image_shape[1]

            x_shape = (next_x_start - previous_x_end - 1)
            self._fill_dask_pieces(dask_pieces, (y_shape, x_shape), chunks)

        return dask_pieces

    @staticmethod
    def _fill_dask_pieces(dask_pieces, shape, chunks):
        if shape[1] > 0:
            new_piece = da.full(shape, np.nan, chunks=chunks)
            dask_pieces.append(new_piece)


def interpolate_slice(slice_rows, slice_cols, interpolator):
    """Interpolate the given slice of the larger array."""
    fine_rows = np.arange(slice_rows.start, slice_rows.stop, slice_rows.step)
    fine_cols = np.arange(slice_cols.start, slice_cols.stop, slice_cols.step)
    return interpolator(fine_cols, fine_rows)


class _AzimuthBlock:
    """Implementation of an single azimuth-noise block."""

    def __init__(self, xml_element):
        """Set up the block from an XML element."""
        self.element = xml_element

    def expand(self, chunks):
        """Build an azimuth block from xml data."""
        corr = 1
        # This isn't needed with newer data (> 2020). When was the change operated?
        #
        #         The azimuth noise is normalized per swath to account for gain
        #         differences between the swaths in EW mode.
        #
        #         This is based on the this reference:
        #         J. Park, A. A. Korosov, M. Babiker, S. Sandven and J. Won,
        #         "Efficient Thermal Noise Removal for Sentinel-1 TOPSAR Cross-Polarization Channel,"
        #         in IEEE Transactions on Geoscience and Remote Sensing, vol. 56, no. 3,
        #         pp. 1555-1565, March 2018.
        #         doi: 10.1109/TGRS.2017.2765248
        #
        # For old data. < 2020
        # swath = elt.find('swath').text
        # if swath == 'EW1':
        #     corr = 1.5
        # if swath in ['EW4', 'IW3']:
        #     corr = 1.2
        # if swath == 'EW5':
        #     corr = 1.5
        data = self.lut * corr

        x_coord = np.arange(self.first_pixel, self.last_pixel + 1)
        y_coord = np.arange(self.first_line, self.last_line + 1)

        new_arr = (da.ones((len(y_coord), len(x_coord)), chunks=chunks) *
                   np.interp(y_coord, self.lines, data)[:, np.newaxis])
        new_arr = xr.DataArray(new_arr,
                               dims=['y', 'x'],
                               coords={'x': x_coord,
                                       'y': y_coord})
        return new_arr

    @property
    def first_pixel(self):
        return int(self.element.find('firstRangeSample').text)

    @property
    def last_pixel(self):
        return int(self.element.find('lastRangeSample').text)

    @property
    def first_line(self):
        return int(self.element.find('firstAzimuthLine').text)

    @property
    def last_line(self):
        return int(self.element.find('lastAzimuthLine').text)

    @property
    def lines(self):
        lines = self.element.find('line').text.split()
        return np.array(lines).astype(int)

    @property
    def lut(self):
        lut = self.element.find('noiseAzimuthLut').text.split()
        return np.array(lut).astype(float)


class XMLArray:
    """A proxy for getting xml data as an array."""

    def __init__(self, root, list_tag, element_tag):
        """Set up the XML array."""
        self.root = root
        self.list_tag = list_tag
        self.element_tag = element_tag
        self.data, self.low_res_coords = self._read_xml_array()

    def expand(self, shape, chunks=None):
        """Generate the full-blown array."""
        return self.interpolate_xml_array(shape, chunks=chunks)

    def _read_xml_array(self):
        """Read an array from xml."""
        elements = self.get_data_items()
        y = []
        x = []
        data = []
        for elt in elements:
            new_x = elt.find('pixel').text.split()
            y += [int(elt.find('line').text)] * len(new_x)
            x += [int(val) for val in new_x]
            data += [float(val)
                     for val in elt.find(self.element_tag).text.split()]

        return np.asarray(data), (x, y)

    def get_data_items(self):
        """Get the data items for this array."""
        data_items = self.root.findall(self.list_tag)
        if not data_items:
            raise KeyError("Can't find data items for xml tag " + self.list_tag)
        return data_items

    def interpolate_xml_array(self, shape, chunks):
        """Interpolate arbitrary size dataset to a full sized grid."""
        xpoints, ypoints = self.low_res_coords
        return interpolate_xarray_linear(xpoints, ypoints, self.data, shape, chunks=chunks)


def interpolate_xarray(xpoints, ypoints, values, shape, kind='cubic',
                       blocksize=CHUNK_SIZE):
    """Interpolate, generating a dask array."""
    vchunks = range(0, shape[0], blocksize)
    hchunks = range(0, shape[1], blocksize)

    token = tokenize(blocksize, xpoints, ypoints, values, kind, shape)
    name = 'interpolate-' + token

    from scipy.interpolate import interp2d
    interpolator = interp2d(xpoints, ypoints, values, kind=kind)

    dskx = {(name, i, j): (interpolate_slice,
                           slice(vcs, min(vcs + blocksize, shape[0])),
                           slice(hcs, min(hcs + blocksize, shape[1])),
                           interpolator)
            for i, vcs in enumerate(vchunks)
            for j, hcs in enumerate(hchunks)
            }

    res = da.Array(dskx, name, shape=list(shape),
                   chunks=(blocksize, blocksize),
                   dtype=values.dtype)
    return DataArray(res, dims=('y', 'x'))


def intp(grid_x, grid_y, interpolator):
    """Interpolate."""
    return interpolator((grid_y, grid_x))


def interpolate_xarray_linear(xpoints, ypoints, values, shape, chunks=CHUNK_SIZE):
    """Interpolate linearly, generating a dask array."""
    from scipy.interpolate.interpnd import LinearNDInterpolator, _ndim_coords_from_arrays

    if isinstance(chunks, (list, tuple)):
        vchunks, hchunks = chunks
    else:
        vchunks, hchunks = chunks, chunks

    points = _ndim_coords_from_arrays(np.vstack((np.asarray(ypoints),
                                                 np.asarray(xpoints))).T)

    interpolator = LinearNDInterpolator(points, values)

    grid_x, grid_y = da.meshgrid(da.arange(shape[1], chunks=hchunks),
                                 da.arange(shape[0], chunks=vchunks))

    # workaround for non-thread-safe first call of the interpolator:
    interpolator((0, 0))
    res = da.map_blocks(intp, grid_x, grid_y, interpolator=interpolator)

    return DataArray(res, dims=('y', 'x'))


class SAFEGRD(BaseFileHandler):
    """Measurement file reader.

    The measurement files are in geotiff format and read using rasterio. For
    performance reasons, the reading adapts the chunk size to match the file's
    block size.
    """

    def __init__(self, filename, filename_info, filetype_info, calfh, noisefh, annotationfh):
        """Init the grd filehandler."""
        super(SAFEGRD, self).__init__(filename, filename_info,
                                      filetype_info)

        self._start_time = filename_info['start_time']
        self._end_time = filename_info['end_time']

        self._polarization = filename_info['polarization']

        self._mission_id = filename_info['mission_id']

        self.calibration = calfh
        self.noise = noisefh
        self.annotation = annotationfh
        self.read_lock = Lock()

        self.filehandle = rasterio.open(self.filename, 'r', sharing=False)

    def get_dataset(self, key, info):
        """Load a dataset."""
        if self._polarization != key["polarization"]:
            return

        logger.debug('Reading %s.', key['name'])

        if key['name'] in ['longitude', 'latitude', 'altitude']:
            logger.debug('Constructing coordinate arrays.')
            arrays = dict()
            arrays['longitude'], arrays['latitude'], arrays['altitude'] = self.get_lonlatalts()

            data = arrays[key['name']]
            data.attrs.update(info)

        else:
            data = rioxarray.open_rasterio(self.filename, lock=False, chunks=(1, CHUNK_SIZE, CHUNK_SIZE)).squeeze()
            data = data.assign_coords(x=np.arange(len(data.coords['x'])),
                                      y=np.arange(len(data.coords['y'])))
            data = self._calibrate_and_denoise(data, key)
            data.attrs.update(info)
            data.attrs.update({'platform_name': self._mission_id})

            data = self._change_quantity(data, key['quantity'])

        return data

    @staticmethod
    def _change_quantity(data, quantity):
        """Change quantity to dB if needed."""
        if quantity == 'dB':
            data.data = 10 * np.log10(data.data)
            data.attrs['units'] = 'dB'
        else:
            data.attrs['units'] = '1'

        return data

    def _calibrate_and_denoise(self, data, key):
        """Calibrate and denoise the data."""
        chunks = CHUNK_SIZE

        dn = self._get_digital_number(data)
        dn = self._denoise(dn, chunks)
        data = self._calibrate(dn, chunks, key)

        return data

    def _get_digital_number(self, data):
        """Get the digital numbers (uncalibrated data)."""
        data = data.where(data > 0)
        data = data.astype(np.float64)
        dn = data * data
        return dn

    def _denoise(self, dn, chunks):
        """Denoise the data."""
        logger.debug('Reading noise data.')
        noise = self.noise.get_noise_correction(chunks=chunks).fillna(0)
        dn = dn - noise
        return dn

    def _calibrate(self, dn, chunks, key):
        """Calibrate the data."""
        logger.debug('Reading calibration data.')
        cal = self.calibration.get_calibration(key['calibration'], chunks=chunks)
        cal_constant = self.calibration.get_calibration_constant()
        logger.debug('Calibrating.')
        data = ((dn + cal_constant) / (cal ** 2)).clip(min=0)
        return data

    @lru_cache(maxsize=2)
    def get_lonlatalts(self):
        """Obtain GCPs and construct latitude and longitude arrays.

        Args:
           band (gdal band): Measurement band which comes with GCP's
           array_shape (tuple) : The size of the data array
        Returns:
           coordinates (tuple): A tuple with longitude and latitude arrays
        """
        band = self.filehandle

        (xpoints, ypoints), (gcp_lons, gcp_lats, gcp_alts), (gcps, crs) = self.get_gcps()

        # FIXME: do interpolation on cartesian coordinates if the area is
        # problematic.

        longitudes = interpolate_xarray(xpoints, ypoints, gcp_lons, band.shape)
        latitudes = interpolate_xarray(xpoints, ypoints, gcp_lats, band.shape)
        altitudes = interpolate_xarray(xpoints, ypoints, gcp_alts, band.shape)

        longitudes.attrs['gcps'] = gcps
        longitudes.attrs['crs'] = crs
        latitudes.attrs['gcps'] = gcps
        latitudes.attrs['crs'] = crs
        altitudes.attrs['gcps'] = gcps
        altitudes.attrs['crs'] = crs

        return longitudes, latitudes, altitudes

    def get_gcps(self):
        """Read GCP from the GDAL band.

        Args:
           band (gdal band): Measurement band which comes with GCP's
           coordinates (tuple): A tuple with longitude and latitude arrays

        Returns:
           points (tuple): Pixel and Line indices 1d arrays
           gcp_coords (tuple): longitude and latitude 1d arrays

        """
        gcps = self.filehandle.gcps

        gcp_array = np.array([(p.row, p.col, p.x, p.y, p.z) for p in gcps[0]])

        ypoints = np.unique(gcp_array[:, 0])
        xpoints = np.unique(gcp_array[:, 1])

        gcp_lons = gcp_array[:, 2].reshape(ypoints.shape[0], xpoints.shape[0])
        gcp_lats = gcp_array[:, 3].reshape(ypoints.shape[0], xpoints.shape[0])
        gcp_alts = gcp_array[:, 4].reshape(ypoints.shape[0], xpoints.shape[0])

        return (xpoints, ypoints), (gcp_lons, gcp_lats, gcp_alts), gcps

    @property
    def start_time(self):
        """Get the start time."""
        return self._start_time

    @property
    def end_time(self):
        """Get the end time."""
        return self._end_time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""HRIT/LRIT format reader.

This module is the base module for all HRIT-based formats. Here, you will find
the common building blocks for hrit reading.

One of the features here is the on-the-fly decompression of hrit files. It needs
a path to the xRITDecompress binary to be provided through the environment
variable called XRIT_DECOMPRESS_PATH. When compressed hrit files are then
encountered (files finishing with `.C_`), they are decompressed to the system's
temporary directory for reading.

"""

import logging
import os
from datetime import timedelta
from io import BytesIO
from subprocess import PIPE, Popen
from tempfile import gettempdir

import dask.array as da
import numpy as np
import xarray as xr
from pyresample import geometry

from satpy.readers.eum_base import time_cds_short
from satpy.readers.file_handlers import BaseFileHandler
from satpy.readers.seviri_base import dec10216

logger = logging.getLogger('hrit_base')

common_hdr = np.dtype([('hdr_id', 'u1'),
                       ('record_length', '>u2')])

primary_header = np.dtype([('file_type', 'u1'),
                           ('total_header_length', '>u4'),
                           ('data_field_length', '>u8')])

image_structure = np.dtype([('number_of_bits_per_pixel', 'u1'),
                            ('number_of_columns', '>u2'),
                            ('number_of_lines', '>u2'),
                            ('compression_flag_for_data', 'u1')])

image_navigation = np.dtype([('projection_name', 'S32'),
                             ('cfac', '>i4'),
                             ('lfac', '>i4'),
                             ('coff', '>i4'),
                             ('loff', '>i4')])

image_data_function = np.dtype([('function', '|S1')])

annotation_header = np.dtype([('annotation', '|S1')])

timestamp_record = np.dtype([('cds_p_field', 'u1'),
                             ('timestamp', time_cds_short)])

ancillary_text = np.dtype([('ancillary', '|S1')])

key_header = np.dtype([('key', '|S1')])

base_text_headers = {image_data_function: 'image_data_function',
                     annotation_header: 'annotation_header',
                     ancillary_text: 'ancillary_text',
                     key_header: 'key_header'}

base_hdr_map = {0: primary_header,
                1: image_structure,
                2: image_navigation,
                3: image_data_function,
                4: annotation_header,
                5: timestamp_record,
                6: ancillary_text,
                7: key_header,
                }


def get_xritdecompress_cmd():
    """Find a valid binary for the xRITDecompress command."""
    cmd = os.environ.get('XRIT_DECOMPRESS_PATH', None)
    if not cmd:
        raise IOError("XRIT_DECOMPRESS_PATH is not defined (complete path to xRITDecompress)")

    question = ("Did you set the environment variable XRIT_DECOMPRESS_PATH correctly?")
    if not os.path.exists(cmd):
        raise IOError(str(cmd) + " does not exist!\n" + question)
    elif os.path.isdir(cmd):
        raise IOError(str(cmd) + " is a directory!\n" + question)

    return cmd


def get_xritdecompress_outfile(stdout):
    """Analyse the output of the xRITDecompress command call and return the file."""
    outfile = b''
    for line in stdout:
        try:
            k, v = [x.strip() for x in line.split(b':', 1)]
        except ValueError:
            break
        if k == b'Decompressed file':
            outfile = v
            break

    return outfile


def decompress(infile, outdir='.'):
    """Decompress an XRIT data file and return the path to the decompressed file.

    It expect to find Eumetsat's xRITDecompress through the environment variable
    XRIT_DECOMPRESS_PATH.
    """
    cmd = get_xritdecompress_cmd()
    infile = os.path.abspath(infile)
    cwd = os.getcwd()
    os.chdir(outdir)

    p = Popen([cmd, infile], stdout=PIPE)
    stdout = BytesIO(p.communicate()[0])
    status = p.returncode
    os.chdir(cwd)

    if status != 0:
        raise IOError("xrit_decompress '%s', failed, status=%d" % (infile, status))

    outfile = get_xritdecompress_outfile(stdout)

    if not outfile:
        raise IOError("xrit_decompress '%s', failed, no output file is generated" % infile)

    return os.path.join(outdir, outfile.decode('utf-8'))


class HRITFileHandler(BaseFileHandler):
    """HRIT standard format reader."""

    def __init__(self, filename, filename_info, filetype_info, hdr_info):
        """Initialize the reader."""
        super(HRITFileHandler, self).__init__(filename, filename_info,
                                              filetype_info)

        self.mda = {}
        self._get_hd(hdr_info)

        if self.mda.get('compression_flag_for_data'):
            logger.debug('Unpacking %s', filename)
            try:
                self.filename = decompress(filename, gettempdir())
            except IOError as err:
                logger.warning("Unpacking failed: %s", str(err))
            self.mda = {}
            self._get_hd(hdr_info)

        self._start_time = filename_info['start_time']
        self._end_time = self._start_time + timedelta(minutes=15)

    def _get_hd(self, hdr_info):
        """Open the file, read and get the basic file header info and set the mda dictionary."""
        hdr_map, variable_length_headers, text_headers = hdr_info

        with open(self.filename) as fp:
            total_header_length = 16
            while fp.tell() < total_header_length:
                hdr_id = np.fromfile(fp, dtype=common_hdr, count=1)[0]
                the_type = hdr_map[hdr_id['hdr_id']]
                if the_type in variable_length_headers:
                    field_length = int((hdr_id['record_length'] - 3) /
                                       the_type.itemsize)
                    current_hdr = np.fromfile(fp,
                                              dtype=the_type,
                                              count=field_length)
                    key = variable_length_headers[the_type]
                    if key in self.mda:
                        if not isinstance(self.mda[key], list):
                            self.mda[key] = [self.mda[key]]
                        self.mda[key].append(current_hdr)
                    else:
                        self.mda[key] = current_hdr
                elif the_type in text_headers:
                    field_length = int((hdr_id['record_length'] - 3) /
                                       the_type.itemsize)
                    char = list(the_type.fields.values())[0][0].char
                    new_type = np.dtype(char + str(field_length))
                    current_hdr = np.fromfile(fp,
                                              dtype=new_type,
                                              count=1)[0]
                    self.mda[text_headers[the_type]] = current_hdr
                else:
                    current_hdr = np.fromfile(fp,
                                              dtype=the_type,
                                              count=1)[0]
                    self.mda.update(
                        dict(zip(current_hdr.dtype.names, current_hdr)))

                total_header_length = self.mda['total_header_length']

        self.mda.setdefault('number_of_bits_per_pixel', 10)

        self.mda['projection_parameters'] = {'a': 6378169.00,
                                             'b': 6356583.80,
                                             'h': 35785831.00,
                                             # FIXME: find a reasonable SSP
                                             'SSP_longitude': 0.0}
        self.mda['orbital_parameters'] = {}

    def get_shape(self, dsid, ds_info):
        """Get shape."""
        return int(self.mda['number_of_lines']), int(self.mda['number_of_columns'])

    @property
    def start_time(self):
        """Get start time."""
        return self._start_time

    @property
    def end_time(self):
        """Get end time."""
        return self._end_time

    def get_dataset(self, key, info):
        """Load a dataset."""
        # Read bands
        data = self.read_band(key, info)

        # Convert to xarray
        xdata = xr.DataArray(data, dims=['y', 'x'])

        return xdata

    def get_xy_from_linecol(self, line, col, offsets, factors):
        """Get the intermediate coordinates from line & col.

        Intermediate coordinates are actually the instruments scanning angles.
        """
        loff, coff = offsets
        lfac, cfac = factors
        x__ = (col - coff) / cfac * 2**16
        y__ = (line - loff) / lfac * 2**16

        return x__, y__

    def get_area_extent(self, size, offsets, factors, platform_height):
        """Get the area extent of the file."""
        nlines, ncols = size
        h = platform_height

        # count starts at 1
        cols = 1 - 0.5
        lines = 1 - 0.5
        ll_x, ll_y = self.get_xy_from_linecol(lines, cols, offsets, factors)

        cols += ncols
        lines += nlines
        ur_x, ur_y = self.get_xy_from_linecol(lines, cols, offsets, factors)

        return (np.deg2rad(ll_x) * h, np.deg2rad(ll_y) * h,
                np.deg2rad(ur_x) * h, np.deg2rad(ur_y) * h)

    def get_area_def(self, dsid):
        """Get the area definition of the band."""
        cfac = np.int32(self.mda['cfac'])
        lfac = np.int32(self.mda['lfac'])
        coff = np.float32(self.mda['coff'])
        loff = np.float32(self.mda['loff'])

        a = self.mda['projection_parameters']['a']
        b = self.mda['projection_parameters']['b']
        h = self.mda['projection_parameters']['h']
        lon_0 = self.mda['projection_parameters']['SSP_longitude']
        nlines = int(self.mda['number_of_lines'])
        ncols = int(self.mda['number_of_columns'])

        area_extent = self.get_area_extent((nlines, ncols),
                                           (loff, coff),
                                           (lfac, cfac),
                                           h)

        proj_dict = {'a': float(a),
                     'b': float(b),
                     'lon_0': float(lon_0),
                     'h': float(h),
                     'proj': 'geos',
                     'units': 'm'}

        area = geometry.AreaDefinition(
            'some_area_name',
            "On-the-fly area",
            'geosmsg',
            proj_dict,
            ncols,
            nlines,
            area_extent)

        self.area = area
        return area

    def read_band(self, key, info):
        """Read the data."""
        shape = int(np.ceil(self.mda['data_field_length'] / 8.))
        if self.mda['number_of_bits_per_pixel'] == 16:
            dtype = '>u2'
            shape //= 2
        elif self.mda['number_of_bits_per_pixel'] in [8, 10]:
            dtype = np.uint8
        shape = (shape, )
        data = np.memmap(self.filename, mode='r',
                         offset=self.mda['total_header_length'],
                         dtype=dtype,
                         shape=shape)
        data = da.from_array(data, chunks=shape[0])
        if self.mda['number_of_bits_per_pixel'] == 10:
            data = dec10216(data)
        data = data.reshape((self.mda['number_of_lines'],
                             self.mda['number_of_columns']))
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The ami_l1b reader tests package."""

import unittest
from unittest import mock

import dask.array as da
import numpy as np
import xarray as xr


class FakeDataset(object):
    """Mimic xarray Dataset object."""

    def __init__(self, info, attrs):
        """Initialize test data."""
        for var_name, var_data in list(info.items()):
            if isinstance(var_data, np.ndarray):
                info[var_name] = xr.DataArray(var_data)
        self.info = info
        self.attrs = attrs

    def __getitem__(self, key):
        """Mimic getitem method."""
        return self.info[key]

    def __contains__(self, key):
        """Mimic contains method."""
        return key in self.info

    def rename(self, *args, **kwargs):
        """Mimic rename method."""
        return self

    def close(self):
        """Act like close method."""
        return


class TestAMIL1bNetCDFBase(unittest.TestCase):
    """Common setup for NC_ABI_L1B tests."""

    @mock.patch('satpy.readers.ami_l1b.xr')
    def setUp(self, xr_, counts=None):
        """Create a fake dataset using the given counts data."""
        from satpy.readers.ami_l1b import AMIL1bNetCDF

        if counts is None:
            rad_data = (np.arange(10.).reshape((2, 5)) + 1.) * 50.
            rad_data = (rad_data + 1.) / 0.5
            rad_data = rad_data.astype(np.int16)
            counts = xr.DataArray(
                da.from_array(rad_data, chunks='auto'),
                dims=('y', 'x'),
                attrs={
                    'channel_name': "VI006",
                    'detector_side': 2,
                    'number_of_total_pixels': 484000000,
                    'number_of_error_pixels': 113892451,
                    'max_pixel_value': 32768,
                    'min_pixel_value': 6,
                    'average_pixel_value': 8228.98770845248,
                    'stddev_pixel_value': 13621.130386551,
                    'number_of_total_bits_per_pixel': 16,
                    'number_of_data_quality_flag_bits_per_pixel': 2,
                    'number_of_valid_bits_per_pixel': 12,
                    'data_quality_flag_meaning':
                        "0:good_pixel, 1:conditionally_usable_pixel, 2:out_of_scan_area_pixel, 3:error_pixel",
                    'ground_sample_distance_ew': 1.4e-05,
                    'ground_sample_distance_ns': 1.4e-05,
                }
            )
        sc_position = xr.DataArray(0., attrs={
            'sc_position_center_pixel': [-26113466.1974016, 33100139.1630508, 3943.75470244799],
        })
        xr_.open_dataset.return_value = FakeDataset(
            {
                'image_pixel_values': counts,
                'sc_position': sc_position,
                'gsics_coeff_intercept': [0.1859369],
                'gsics_coeff_slope': [0.9967594],
            },
            {
                "satellite_name": "GK-2A",
                "observation_start_time": 623084431.957882,
                "observation_end_time": 623084975.606133,
                "projection_type": "GEOS",
                "sub_longitude": 2.23751210105673,
                "cfac": 81701355.6133574,
                "lfac": -81701355.6133574,
                "coff": 11000.5,
                "loff": 11000.5,
                "nominal_satellite_height": 42164000.,
                "earth_equatorial_radius": 6378137.,
                "earth_polar_radius": 6356752.3,
                "number_of_columns": 22000,
                "number_of_lines": 22000,
                "observation_mode": "FD",
                "channel_spatial_resolution": "0.5",
                "Radiance_to_Albedo_c": 1,
                "DN_to_Radiance_Gain": -0.0144806550815701,
                "DN_to_Radiance_Offset": 118.050903320312,
                "Teff_to_Tbb_c0": -0.141418528203155,
                "Teff_to_Tbb_c1": 1.00052232906885,
                "Teff_to_Tbb_c2": -0.00000036287276076109,
                "light_speed": 2.9979245800E+08,
                "Boltzmann_constant_k": 1.3806488000E-23,
                "Plank_constant_h": 6.6260695700E-34,
            }
        )

        self.reader = AMIL1bNetCDF('filename',
                                   {'platform_shortname': 'gk2a'},
                                   {'file_type': 'ir087'},)


class TestAMIL1bNetCDF(TestAMIL1bNetCDFBase):
    """Test the AMI L1b reader."""

    def _check_orbital_parameters(self, orb_params):
        """Check that orbital parameters match expected values."""
        exp_params = {
            'projection_altitude': 35785863.0,
            'projection_latitude': 0.0,
            'projection_longitude': 128.2,
            'satellite_actual_altitude': 35782654.56070405,
            'satellite_actual_latitude': 0.005364927,
            'satellite_actual_longitude': 128.2707,
        }
        for key, val in exp_params.items():
            self.assertAlmostEqual(val, orb_params[key], places=3)

    def test_filename_grouping(self):
        """Test that filenames are grouped properly."""
        from satpy.readers import group_files
        filenames = [
            'gk2a_ami_le1b_ir087_fd020ge_201909300300.nc',
            'gk2a_ami_le1b_ir096_fd020ge_201909300300.nc',
            'gk2a_ami_le1b_ir105_fd020ge_201909300300.nc',
            'gk2a_ami_le1b_ir112_fd020ge_201909300300.nc',
            'gk2a_ami_le1b_ir123_fd020ge_201909300300.nc',
            'gk2a_ami_le1b_ir133_fd020ge_201909300300.nc',
            'gk2a_ami_le1b_nr013_fd020ge_201909300300.nc',
            'gk2a_ami_le1b_nr016_fd020ge_201909300300.nc',
            'gk2a_ami_le1b_sw038_fd020ge_201909300300.nc',
            'gk2a_ami_le1b_vi004_fd010ge_201909300300.nc',
            'gk2a_ami_le1b_vi005_fd010ge_201909300300.nc',
            'gk2a_ami_le1b_vi006_fd005ge_201909300300.nc',
            'gk2a_ami_le1b_vi008_fd010ge_201909300300.nc',
            'gk2a_ami_le1b_wv063_fd020ge_201909300300.nc',
            'gk2a_ami_le1b_wv069_fd020ge_201909300300.nc',
            'gk2a_ami_le1b_wv073_fd020ge_201909300300.nc']
        groups = group_files(filenames, reader='ami_l1b')
        self.assertEqual(len(groups), 1)
        self.assertEqual(len(groups[0]['ami_l1b']), 16)

    def test_basic_attributes(self):
        """Test getting basic file attributes."""
        from datetime import datetime
        self.assertEqual(self.reader.start_time,
                         datetime(2019, 9, 30, 3, 0, 31, 957882))
        self.assertEqual(self.reader.end_time,
                         datetime(2019, 9, 30, 3, 9, 35, 606133))

    def test_get_dataset(self):
        """Test gettting radiance data."""
        from satpy.tests.utils import make_dataid
        key = make_dataid(name='VI006', calibration='radiance')
        res = self.reader.get_dataset(key, {
            'file_key': 'image_pixel_values',
            'standard_name': 'toa_outgoing_radiance_per_unit_wavelength',
            'units': 'W m-2 um-1 sr-1',
        })
        exp = {'calibration': 'radiance',
               'modifiers': (),
               'platform_name': 'GEO-KOMPSAT-2A',
               'sensor': 'ami',
               'units': 'W m-2 um-1 sr-1'}
        for key, val in exp.items():
            self.assertEqual(val, res.attrs[key])
        self._check_orbital_parameters(res.attrs['orbital_parameters'])

    def test_bad_calibration(self):
        """Test that asking for a bad calibration fails."""
        from satpy.tests.utils import make_dataid
        with self.assertRaises(ValueError):
            ds_id = make_dataid(name='VI006', calibration='_bad_')
            ds_info = {'file_key': 'image_pixel_values',
                       'standard_name': 'toa_outgoing_radiance_per_unit_wavelength',
                       'units': 'W m-2 um-1 sr-1',
                       }
            self.reader.get_dataset(ds_id, ds_info)

    @mock.patch('satpy.readers.abi_base.geometry.AreaDefinition')
    def test_get_area_def(self, adef):
        """Test the area generation."""
        self.reader.get_area_def(None)

        self.assertEqual(adef.call_count, 1)
        call_args = tuple(adef.call_args)[0]
        exp = {'a': 6378137.0, 'b': 6356752.3, 'h': 35785863.0,
               'lon_0': 128.2, 'proj': 'geos', 'units': 'm'}
        for key, val in exp.items():
            self.assertIn(key, call_args[3])
            self.assertAlmostEqual(val, call_args[3][key])
        self.assertEqual(call_args[4], self.reader.nc.attrs['number_of_columns'])
        self.assertEqual(call_args[5], self.reader.nc.attrs['number_of_lines'])
        np.testing.assert_allclose(call_args[6],
                                   [-5511022.902, -5511022.902, 5511022.902, 5511022.902])

    def test_get_dataset_vis(self):
        """Test get visible calibrated data."""
        from satpy.tests.utils import make_dataid
        key = make_dataid(name='VI006', calibration='reflectance')
        res = self.reader.get_dataset(key, {
            'file_key': 'image_pixel_values',
            'standard_name': 'toa_bidirectional_reflectance',
            'units': '%',
        })
        exp = {'calibration': 'reflectance',
               'modifiers': (),
               'platform_name': 'GEO-KOMPSAT-2A',
               'sensor': 'ami',
               'units': '%'}
        for key, val in exp.items():
            self.assertEqual(val, res.attrs[key])
        self._check_orbital_parameters(res.attrs['orbital_parameters'])

    def test_get_dataset_counts(self):
        """Test get counts data."""
        from satpy.tests.utils import make_dataid
        key = make_dataid(name='VI006', calibration='counts')
        res = self.reader.get_dataset(key, {
            'file_key': 'image_pixel_values',
            'standard_name': 'counts',
            'units': '1',
        })
        exp = {'calibration': 'counts',
               'modifiers': (),
               'platform_name': 'GEO-KOMPSAT-2A',
               'sensor': 'ami',
               'units': '1'}
        for key, val in exp.items():
            self.assertEqual(val, res.attrs[key])
        self._check_orbital_parameters(res.attrs['orbital_parameters'])


class TestAMIL1bNetCDFIRCal(TestAMIL1bNetCDFBase):
    """Test IR specific things about the AMI reader."""

    def setUp(self):
        """Create test data for IR calibration tests."""
        from satpy.tests.utils import make_dataid
        count_data = (np.arange(10).reshape((2, 5))) + 7000
        count_data = count_data.astype(np.uint16)
        count = xr.DataArray(
            da.from_array(count_data, chunks='auto'),
            dims=('y', 'x'),
            attrs={
                'channel_name': "IR087",
                'detector_side': 2,
                'number_of_total_pixels': 484000000,
                'number_of_error_pixels': 113892451,
                'max_pixel_value': 32768,
                'min_pixel_value': 6,
                'average_pixel_value': 8228.98770845248,
                'stddev_pixel_value': 13621.130386551,
                'number_of_total_bits_per_pixel': 16,
                'number_of_data_quality_flag_bits_per_pixel': 2,
                'number_of_valid_bits_per_pixel': 13,
                'data_quality_flag_meaning':
                    "0:good_pixel, 1:conditionally_usable_pixel, 2:out_of_scan_area_pixel, 3:error_pixel",
                'ground_sample_distance_ew': 1.4e-05,
                'ground_sample_distance_ns': 1.4e-05,
            }
        )
        self.ds_id = make_dataid(name='IR087', wavelength=[8.415, 8.59, 8.765],
                                 calibration='brightness_temperature')
        self.ds_info = {
            'file_key': 'image_pixel_values',
            'wavelength': [8.415, 8.59, 8.765],
            'standard_name': 'toa_brightness_temperature',
            'units': 'K',
        }
        super(TestAMIL1bNetCDFIRCal, self).setUp(counts=count)

    def test_default_calibrate(self):
        """Test default (pyspectral) IR calibration."""
        from satpy.readers.ami_l1b import rad2temp
        with mock.patch('satpy.readers.ami_l1b.rad2temp', wraps=rad2temp) as r2t_mock:
            res = self.reader.get_dataset(self.ds_id, self.ds_info)
            r2t_mock.assert_called_once()
        expected = np.array([[238.34385135, 238.31443527, 238.28500087, 238.25554813, 238.22607701],
                             [238.1965875, 238.16707956, 238.13755317, 238.10800829, 238.07844489]])
        np.testing.assert_allclose(res.data.compute(), expected, equal_nan=True)
        # make sure the attributes from the file are in the data array
        self.assertEqual(res.attrs['standard_name'], 'toa_brightness_temperature')

    def test_infile_calibrate(self):
        """Test IR calibration using in-file coefficients."""
        from satpy.readers.ami_l1b import rad2temp
        self.reader.calib_mode = 'FILE'
        with mock.patch('satpy.readers.ami_l1b.rad2temp', wraps=rad2temp) as r2t_mock:
            res = self.reader.get_dataset(self.ds_id, self.ds_info)
            r2t_mock.assert_not_called()
        expected = np.array([[238.34385135, 238.31443527, 238.28500087, 238.25554813, 238.22607701],
                             [238.1965875, 238.16707956, 238.13755317, 238.10800829, 238.07844489]])
        # file coefficients are pretty close, give some wiggle room
        np.testing.assert_allclose(res.data.compute(), expected, equal_nan=True, atol=0.04)
        # make sure the attributes from the file are in the data array
        self.assertEqual(res.attrs['standard_name'], 'toa_brightness_temperature')

    def test_gsics_radiance_corr(self):
        """Test IR radiance adjustment using in-file GSICS coefs."""
        from satpy.readers.ami_l1b import rad2temp
        self.reader.calib_mode = 'GSICS'
        expected = np.array([[238.036797, 238.007106, 237.977396, 237.947668, 237.91792],
                             [237.888154, 237.85837, 237.828566, 237.798743, 237.768902]])
        with mock.patch('satpy.readers.ami_l1b.rad2temp', wraps=rad2temp) as r2t_mock:
            res = self.reader.get_dataset(self.ds_id, self.ds_info)
            r2t_mock.assert_not_called()
        # file coefficients are pretty close, give some wiggle room
        np.testing.assert_allclose(res.data.compute(), expected, equal_nan=True, atol=0.01)
        # make sure the attributes from the file are in the data array
        self.assertEqual(res.attrs['standard_name'], 'toa_brightness_temperature')

    def test_user_radiance_corr(self):
        """Test IR radiance adjustment using user-supplied coefs."""
        from satpy.readers.ami_l1b import rad2temp
        self.reader.calib_mode = 'FILE'
        self.reader.user_calibration = {'IR087': {'slope': 0.99669,
                                                  'offset': 0.16907}}
        expected = np.array([[238.073713, 238.044043, 238.014354, 237.984647, 237.954921],
                             [237.925176, 237.895413, 237.865631, 237.835829, 237.806009]])
        with mock.patch('satpy.readers.ami_l1b.rad2temp', wraps=rad2temp) as r2t_mock:
            res = self.reader.get_dataset(self.ds_id, self.ds_info)
            r2t_mock.assert_not_called()
        # file coefficients are pretty close, give some wiggle room
        np.testing.assert_allclose(res.data.compute(), expected, equal_nan=True, atol=0.01)
        # make sure the attributes from the file are in the data array
        self.assertEqual(res.attrs['standard_name'], 'toa_brightness_temperature')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'cmsaf-claas2_l2_nc' reader."""

import datetime
import os
from unittest import mock

import numpy as np
import numpy.testing
import pytest
import xarray as xr

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler


class FakeNetCDF4FileHandler2(FakeNetCDF4FileHandler):
    """Class for faking the NetCDF4 Filehandler."""

    _nrows = 30
    _ncols = 40

    def __init__(self, *args, auto_maskandscale, **kwargs):
        """Init the file handler."""
        # make sure that CLAAS2 reader asks NetCDF4FileHandler for having
        # auto_maskandscale enabled
        assert auto_maskandscale
        super().__init__(*args, **kwargs)

    def _get_global_attributes(self):
        data = {}
        attrs = {
                "CMSAF_proj4_params": "+a=6378169.0 +h=35785831.0 "
                                      "+b=6356583.8 +lon_0=0 +proj=geos",
                "CMSAF_area_extent": np.array(
                    [-5456233.41938636, -5453233.01608472,
                     5453233.01608472, 5456233.41938636]),
                "time_coverage_start": "1985-08-13T13:15:00Z",
                "time_coverage_end": "2085-08-13T13:15:00Z",
                }
        for (k, v) in attrs.items():
            data["/attr/" + k] = v
        return data

    def _get_data(self):
        data = {
                "cph": xr.DataArray(
                    np.arange(self._nrows*self._ncols, dtype="i4").reshape(
                        (1, self._nrows, self._ncols))/100,
                    dims=("time", "y", "x")),
                "ctt": xr.DataArray(
                    np.arange(self._nrows*self._ncols, 0, -1,
                              dtype="i4").reshape(
                                  (self._nrows, self._ncols))/100,
                    dims=("y", "x")),
                "time_bnds": xr.DataArray(
                    [[12436.91666667, 12436.92534722]],
                    dims=("time", "time_bnds"))}
        for k in set(data.keys()):
            data[f"{k:s}/dimensions"] = data[k].dims
            data[f"{k:s}/attr/fruit"] = "apple"
            data[f"{k:s}/attr/scale_factor"] = np.float32(0.01)
        return data

    def _get_dimensions(self):
        data = {
                "/dimension/x": self._nrows,
                "/dimension/y": self._ncols,
                "/dimension/time": 1,
                "/dimension/time_bnds": 2,
                }
        return data

    def get_test_content(self, filename, filename_info, filetype_info):
        """Get the content of the test data."""
        # mock global attributes
        # - root groups global
        # - other groups global
        # mock data variables
        # mock dimensions
        #
        # ... but only what satpy is using ...

        D = {}
        D.update(self._get_data())
        D.update(self._get_dimensions())
        D.update(self._get_global_attributes())
        return D


@pytest.fixture
def reader():
    """Return reader for CMSAF CLAAS-2."""
    from satpy._config import config_search_paths
    from satpy.readers import load_reader

    reader_configs = config_search_paths(
        os.path.join("readers", "cmsaf-claas2_l2_nc.yaml"))
    reader = load_reader(reader_configs)
    return reader


@pytest.fixture(autouse=True, scope="class")
def fake_handler():
    """Wrap NetCDF4 FileHandler with our own fake handler."""
    # implementation strongly inspired by test_viirs_l1b.py
    from satpy.readers.cmsaf_claas2 import CLAAS2
    p = mock.patch.object(
            CLAAS2,
            "__bases__",
            (FakeNetCDF4FileHandler2,))
    with p:
        p.is_local = True
        yield p


def test_file_pattern(reader):
    """Test file pattern matching."""
    filenames = [
            "CTXin20040120091500305SVMSG01MD.nc",
            "CTXin20040120093000305SVMSG01MD.nc",
            "CTXin20040120094500305SVMSG01MD.nc",
            "abcde52034294023489248MVSSG03DD.nc"]

    files = reader.select_files_from_pathnames(filenames)
    # only 3 out of 4 above should match
    assert len(files) == 3


def test_load(reader):
    """Test loading."""
    from satpy.tests.utils import make_dataid

    # testing two filenames to test correctly combined
    filenames = [
        "CTXin20040120091500305SVMSG01MD.nc",
        "CTXin20040120093000305SVMSG01MD.nc"]

    loadables = reader.select_files_from_pathnames(filenames)
    reader.create_filehandlers(loadables)
    res = reader.load(
            [make_dataid(name=name) for name in ["cph", "ctt"]])
    assert 2 == len(res)
    assert reader.start_time == datetime.datetime(1985, 8, 13, 13, 15)
    assert reader.end_time == datetime.datetime(2085, 8, 13, 13, 15)
    np.testing.assert_array_almost_equal(
            res["cph"].data,
            np.tile(np.arange(0.0, 12.0, 0.01).reshape((30, 40)), [2, 1]))
    np.testing.assert_array_almost_equal(
            res["ctt"].data,
            np.tile(np.arange(12.0, 0.0, -0.01).reshape((30, 40)), [2, 1]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the VIIRS EDR Flood reader."""
import os
import unittest
from unittest import mock

import numpy as np

from satpy.tests.reader_tests.test_hdf4_utils import FakeHDF4FileHandler

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)


class FakeHDF4FileHandler2(FakeHDF4FileHandler):
    """Swap in HDF4 file handler."""

    def get_test_content(self, filename, filename_info, filename_type):
        """Mimic reader input file content."""
        file_content = {}
        file_content['/attr/Satellitename'] = filename_info['platform_shortname']
        file_content['/attr/SensorIdentifyCode'] = 'VIIRS'

        # only one dataset for the flood reader
        file_content['WaterDetection'] = DEFAULT_FILE_DATA
        file_content['WaterDetection/attr/_Fillvalue'] = 1
        file_content['WaterDetection/attr/scale_factor'] = 1.
        file_content['WaterDetection/attr/add_offset'] = 0.
        file_content['WaterDetection/attr/units'] = 'none'
        file_content['WaterDetection/shape'] = DEFAULT_FILE_SHAPE
        file_content['WaterDetection/attr/ProjectionMinLatitude'] = 15.
        file_content['WaterDetection/attr/ProjectionMaxLatitude'] = 68.
        file_content['WaterDetection/attr/ProjectionMinLongitude'] = -124.
        file_content['WaterDetection/attr/ProjectionMaxLongitude'] = -61.

        # convert tp xarrays
        from xarray import DataArray
        for key, val in file_content.items():
            if isinstance(val, np.ndarray):
                attrs = {}
                for a in ['_Fillvalue', 'units', 'ProjectionMinLatitude', 'ProjectionMaxLongitude',
                          'ProjectionMinLongitude', 'ProjectionMaxLatitude']:
                    if key + '/attr/' + a in file_content:
                        attrs[a] = file_content[key + '/attr/' + a]
                if val.ndim > 1:
                    file_content[key] = DataArray(val, dims=('fakeDim0', 'fakeDim1'), attrs=attrs)
                else:
                    file_content[key] = DataArray(val, attrs=attrs)

        if 'y' not in file_content['WaterDetection'].dims:
            file_content['WaterDetection'] = file_content['WaterDetection'].rename({'fakeDim0': 'x', 'fakeDim1': 'y'})
        return file_content


class TestVIIRSEDRFloodReader(unittest.TestCase):
    """Test VIIRS EDR Flood Reader."""

    yaml_file = 'viirs_edr_flood.yaml'

    def setUp(self):
        """Wrap HDF4 file handler with own fake file handler."""
        from satpy._config import config_search_paths
        from satpy.readers.viirs_edr_flood import VIIRSEDRFlood
        self.reader_configs = config_search_paths(os.path.join('readers', self.yaml_file))
        self.p = mock.patch.object(VIIRSEDRFlood, '__bases__', (FakeHDF4FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'WATER_VIIRS_Prj_SVI_npp_d20180824_t1828213_e1839433_b35361_cspp_dev_10_300_01.hdf'
        ])
        self.assertEqual(len(loadables), 1)
        r.create_filehandlers(loadables)
        self.assertTrue(r.file_handlers)

    def test_load_dataset(self):
        """Test loading all datasets from a full swath file."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'WATER_VIIRS_Prj_SVI_npp_d20180824_t1828213_e1839433_b35361_cspp_dev_10_300_01.hdf'
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['WaterDetection'])
        self.assertEqual(len(datasets), 1)
        for v in datasets.values():
            self.assertEqual(v.attrs['units'], 'none')

    def test_load_dataset_aoi(self):
        """Test loading all datasets from an area of interest file."""
        from satpy.readers import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            'WATER_VIIRS_Prj_SVI_npp_d20180824_t1828213_e1839433_b35361_cspp_dev_001_10_300_01.hdf'
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(['WaterDetection'])
        self.assertEqual(len(datasets), 1)
        for v in datasets.values():
            self.assertEqual(v.attrs['units'], 'none')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.nc_slstr module."""
import unittest
import unittest.mock as mock
import warnings
from datetime import datetime

import numpy as np
import xarray as xr

from satpy.dataset.dataid import DataID, ModifierTuple, WavelengthRange
from satpy.readers.slstr_l1b import NCSLSTR1B, NCSLSTRAngles, NCSLSTRFlag, NCSLSTRGeo

local_id_keys_config = {'name': {
    'required': True,
},
    'wavelength': {
        'type': WavelengthRange,
    },
    'resolution': None,
    'calibration': {
        'enum': [
            'reflectance',
            'brightness_temperature',
            'radiance',
            'counts'
        ]
    },
    'stripe': {
        'enum': [
            'a',
            'b',
            'c',
            'i',
            'f',
        ]
    },
    'view': {
        'enum': [
            'nadir',
            'oblique',
        ]
    },
    'modifiers': {
        'required': True,
        'default': ModifierTuple(),
        'type': ModifierTuple,
    },
}


class TestSLSTRL1B(unittest.TestCase):
    """Common setup for SLSTR_L1B tests."""

    @mock.patch('satpy.readers.slstr_l1b.xr')
    def setUp(self, xr_):
        """Create a fake dataset using the given radiance data."""
        self.base_data = np.array(([1., 2., 3.], [4., 5., 6.]))
        self.det_data = np.array(([0, 1, 1], [0, 1, 0]))
        self.start_time = "2020-05-10T12:01:15.585Z"
        self.end_time = "2020-05-10T12:06:18.012Z"
        self.rad = xr.DataArray(
            self.base_data,
            dims=('columns', 'rows'),
            attrs={'scale_factor': 1.0, 'add_offset': 0.0,
                   '_FillValue': -32768, 'units': 'mW.m-2.sr-1.nm-1',
                   }
        )
        det = xr.DataArray(
            self.base_data,
            dims=('columns', 'rows'),
            attrs={'scale_factor': 1.0, 'add_offset': 0.0,
                   '_FillValue': 255,
                   }
        )
        self.fake_dataset = xr.Dataset(
            data_vars={
                'S5_radiance_an': self.rad,
                'S9_BT_ao': self.rad,
                'foo_radiance_an': self.rad,
                'S5_solar_irradiances': self.rad,
                'geometry_tn': self.rad,
                'latitude_an': self.rad,
                'x_tx': self.rad,
                'y_tx': self.rad,
                'x_in': self.rad,
                'y_in': self.rad,
                'x_an': self.rad,
                'y_an': self.rad,
                'flags_an': self.rad,
                'detector_an': det,
            },
            attrs={
                "start_time": self.start_time,
                "stop_time": self.end_time,
            },
        )


def make_dataid(**items):
    """Make a data id."""
    return DataID(local_id_keys_config, **items)


class TestSLSTRReader(TestSLSTRL1B):
    """Test various nc_slstr file handlers."""

    class FakeSpl:
        """Fake return function for SPL interpolation."""

        @staticmethod
        def ev(foo_x, foo_y):
            """Fake function to return interpolated data."""
            return np.zeros((3, 2))

    @mock.patch('satpy.readers.slstr_l1b.xr')
    @mock.patch('scipy.interpolate.RectBivariateSpline')
    def test_instantiate(self, bvs_, xr_):
        """Test initialization of file handlers."""
        bvs_.return_value = self.FakeSpl
        xr_.open_dataset.return_value = self.fake_dataset

        good_start = datetime.strptime(self.start_time,
                                       '%Y-%m-%dT%H:%M:%S.%fZ')
        good_end = datetime.strptime(self.end_time,
                                     '%Y-%m-%dT%H:%M:%S.%fZ')

        ds_id = make_dataid(name='foo', calibration='radiance',
                            stripe='a', view='nadir')
        ds_id_500 = make_dataid(name='foo', calibration='radiance',
                                stripe='a', view='nadir', resolution=500)
        filename_info = {'mission_id': 'S3A', 'dataset_name': 'foo',
                         'start_time': 0, 'end_time': 0,
                         'stripe': 'a', 'view': 'n'}
        test = NCSLSTR1B('somedir/S1_radiance_an.nc', filename_info, 'c')
        assert test.view == 'nadir'
        assert test.stripe == 'a'
        test.get_dataset(ds_id, dict(filename_info, **{'file_key': 'foo'}))
        self.assertEqual(test.start_time, good_start)
        self.assertEqual(test.end_time, good_end)
        xr_.open_dataset.assert_called()
        xr_.open_dataset.reset_mock()

        filename_info = {'mission_id': 'S3A', 'dataset_name': 'foo',
                         'start_time': 0, 'end_time': 0,
                         'stripe': 'c', 'view': 'o'}
        test = NCSLSTR1B('somedir/S1_radiance_co.nc', filename_info, 'c')
        assert test.view == 'oblique'
        assert test.stripe == 'c'
        test.get_dataset(ds_id, dict(filename_info, **{'file_key': 'foo'}))
        self.assertEqual(test.start_time, good_start)
        self.assertEqual(test.end_time, good_end)
        xr_.open_dataset.assert_called()
        xr_.open_dataset.reset_mock()

        filename_info = {'mission_id': 'S3A', 'dataset_name': 'foo',
                         'start_time': 0, 'end_time': 0,
                         'stripe': 'a', 'view': 'n'}
        test = NCSLSTRGeo('somedir/geometry_an.nc', filename_info, 'c')
        test.get_dataset(ds_id, dict(filename_info, **{'file_key': 'latitude_{stripe:1s}{view:1s}'}))
        self.assertEqual(test.start_time, good_start)
        self.assertEqual(test.end_time, good_end)
        xr_.open_dataset.assert_called()
        xr_.open_dataset.reset_mock()

        test = NCSLSTRFlag('somedir/S1_radiance_an.nc', filename_info, 'c')
        test.get_dataset(ds_id, dict(filename_info, **{'file_key': 'flags_{stripe:1s}{view:1s}'}))
        assert test.view == 'nadir'
        assert test.stripe == 'a'
        self.assertEqual(test.start_time, good_start)
        self.assertEqual(test.end_time, good_end)
        xr_.open_dataset.assert_called()
        xr_.open_dataset.reset_mock()

        test = NCSLSTRAngles('somedir/S1_radiance_an.nc', filename_info, 'c')
        test.get_dataset(ds_id, dict(filename_info, **{'file_key': 'geometry_t{view:1s}'}))
        self.assertEqual(test.start_time, good_start)
        self.assertEqual(test.end_time, good_end)
        xr_.open_dataset.assert_called()
        xr_.open_dataset.reset_mock()
        test.get_dataset(ds_id_500, dict(filename_info, **{'file_key': 'geometry_t{view:1s}'}))


class TestSLSTRCalibration(TestSLSTRL1B):
    """Test the implementation of the calibration factors."""

    @mock.patch('satpy.readers.slstr_l1b.xr')
    def test_radiance_calibration(self, xr_):
        """Test radiance calibration steps."""
        from satpy.readers.slstr_l1b import CHANCALIB_FACTORS
        xr_.open_dataset.return_value = self.fake_dataset

        ds_id = make_dataid(name='foo', calibration='radiance',
                            stripe='a', view='nadir')
        filename_info = {'mission_id': 'S3A', 'dataset_name': 'foo',
                         'start_time': 0, 'end_time': 0,
                         'stripe': 'a', 'view': 'n'}

        test = NCSLSTR1B('somedir/S1_radiance_co.nc', filename_info, 'c')
        # Check warning is raised if we don't have calibration
        with warnings.catch_warnings(record=True) as w:
            test.get_dataset(ds_id, dict(filename_info, **{'file_key': 'foo'}))
            assert issubclass(w[-1].category, UserWarning)

        # Check user calibration is used correctly
        test = NCSLSTR1B('somedir/S1_radiance_co.nc', filename_info, 'c',
                         user_calibration={'foo_nadir': 0.4})
        data = test.get_dataset(ds_id, dict(filename_info, **{'file_key': 'foo'}))
        np.testing.assert_allclose(data.values, self.base_data * 0.4)

        # Check internal calibration is used correctly
        ds_id = make_dataid(name='S5', calibration='radiance', stripe='a', view='nadir')
        filename_info['dataset_name'] = 'S5'
        test = NCSLSTR1B('somedir/S1_radiance_an.nc', filename_info, 'c')
        data = test.get_dataset(ds_id, dict(filename_info, **{'file_key': 'S5'}))
        np.testing.assert_allclose(data.values,
                                   self.base_data * CHANCALIB_FACTORS['S5_nadir'])

    @mock.patch('satpy.readers.slstr_l1b.xr')
    @mock.patch('satpy.readers.slstr_l1b.da')
    def test_reflectance_calibration(self, da_, xr_):
        """Test reflectance calibration."""
        xr_.open_dataset.return_value = self.fake_dataset
        da_.map_blocks.return_value = self.rad / 100.
        filename_info = {'mission_id': 'S3A', 'dataset_name': 'S5',
                         'start_time': 0, 'end_time': 0,
                         'stripe': 'a', 'view': 'n'}
        ds_id = make_dataid(name='S5', calibration='reflectance', stripe='a', view='nadir')
        test = NCSLSTR1B('somedir/S1_radiance_an.nc', filename_info, 'c')
        data = test.get_dataset(ds_id, dict(filename_info, **{'file_key': 'S5'}))
        self.assertEqual(data.units, '%')
        np.testing.assert_allclose(data.values, self.rad * np.pi)

    def test_cal_rad(self):
        """Test the radiance to reflectance converter."""
        rad = np.array([10., 20., 30., 40., 50., 60., 70.])
        didx = np.array([1, 2., 1., 3., 2., 2., 0.])
        solflux = np.array([100., 200., 300., 400.])

        good_rad = np.array([1. / 20., 1. / 15., 3. / 20., 1. / 10., 1. / 6., 2. / 10., 7. / 10.])

        out_rad = NCSLSTR1B._cal_rad(rad, didx, solflux)
        np.testing.assert_allclose(out_rad, good_rad)
